from typing import Optional
from typing import Tuple

import jax.numpy as jnp
import numpy as np
import numpyro
from jax import random
from numpyro.distributions import Exponential
from numpyro.distributions import HalfCauchy
from numpyro.distributions import InverseGamma
from numpyro.distributions import Normal
from numpyro.infer import MCMC
from numpyro.infer import NUTS

from .base import BaseOptimizer


class SBR(BaseOptimizer):
    """
    Sparse Bayesian Regression (SBR) optimizer. This uses the regularised
    horseshoe prior over the SINDy coefficients to achieve sparsification.

    The horseshoe prior contains a "spike" of nonzero probability at the
    origin, and a Student's-T-shaped "slab" of distribution in cases where a
    coefficient is nonzero.


    The SINDy coefficients are set as the posterior means of the MCMC NUTS samples.
    Additional statistics can be computed from the MCMC samples stored in
    the mcmc attribute using e.g. ArviZ.

    This implementation differs from the method described in Hirsh et al. (2021)
    by imposing the error model directly on the derivatives, rather than on the
    states, circumventing the need to integrate the equation to evaluate the
    posterior density. One consequence of this is that the noise standard
    deviation "sigma" is with respect to the derivatives instead of the states
    and hence should not be interpreted.

    TODO: Implement the data-generating model described in eq. 2.4 of Hirsh
    et al. (2021). This could be achieved using the JAX-based solver "diffrax".
    Se discussion in https://github.com/dynamicslab/pysindy/pull/440 for more
    details.

    See the following reference for more details:

        Hirsh, S. M., Barajas-Solano, D. A., & Kutz, J. N. (2021).
        Sparsifying Priors for Bayesian Uncertainty Quantification in
        Model Discovery (arXiv:2107.02107). arXiv. http://arxiv.org/abs/2107.02107

    Parameters
    ----------
    sparsity_coef_tau0 : float, optional (default 0.1)
        Sparsity coefficient for regularised horseshoe hyper-prior. Lower
        value increases the sparsity of the SINDy coefficients.

    slab_shape_nu : float, optional (default 4)
        Along with ``slab_shape_s``, controls tails of nonzero coefficients.
        Specifically, degrees of freedom for Student's-T-shaped slab.
        Higher values decrease excess kurtosis to zero, lower values >= 4
        increase kurtosis to infinity.

    slab_shape_s : float, optional (default 2)
        Along with ``slab_shape_nu``, controls standard deviation of nonzero
        coefficients.

    noise_hyper_lambda : float, optional (default 1)
        Rate hyperparameter for the exponential prior distribution over
        the noise standard deviation.

    num_warmup : int, optional (default 1000)
        Number of warmup (or "burnin") MCMC samples to generate. These are
        discarded before analysis and are not included in the posterior samples.

    num_samples : int, optional (default 5000)
        Number of posterior MCMC samples to generate.

    mcmc_kwargs : dict, optional (default None)
        Instructions for MCMC sampling.
        Keyword arguments are passed to numpyro.infer.MCMC

    Attributes
    ----------
    coef_ : array, shape (n_features,) or (n_targets, n_features)
        Posterior means of the SINDy coefficients.

    mcmc_ : numpyro.infer.MCMC
        Complete traces of the posterior samples.
    """

    def __init__(
        self,
        sparsity_coef_tau0: float = 0.1,
        slab_shape_nu: float = 4,
        slab_shape_s: float = 2,
        noise_hyper_lambda: float = 1,
        num_warmup: int = 1000,
        num_samples: int = 5000,
        mcmc_kwargs: Optional[dict] = None,
        nuts_kwargs: Optional[dict] = None,
        unbias: bool = False,
        **kwargs,
    ):
        if unbias:
            raise ValueError("SBR is incompatible with unbiasing. Set unbias=False")

        super().__init__(unbias=unbias, **kwargs)

        # check that hyperparameters are positive.
        if sparsity_coef_tau0 <= 0:
            raise ValueError("sparsity_coef_tau0 must be positive")
        if slab_shape_nu <= 0:
            raise ValueError("slab_shape_nu must be positive")
        if slab_shape_s <= 0:
            raise ValueError("slab_shape_s must be positive")
        if noise_hyper_lambda <= 0:
            raise ValueError("noise_hyper_lambda must be positive")

        # check that samples are positive integers.
        if not isinstance(num_warmup, int) or num_warmup < 0:
            raise ValueError("num_warmup must be a positive integer")
        if not isinstance(num_samples, int) or num_samples < 0:
            raise ValueError("num_samples must be a positive integer")

        # set the hyperparameters
        self.sparsity_coef_tau0 = sparsity_coef_tau0
        self.slab_shape_nu = slab_shape_nu
        self.slab_shape_s = slab_shape_s
        self.noise_hyper_lambda = noise_hyper_lambda

        # set MCMC sampling parameters.
        self.num_warmup = num_warmup
        self.num_samples = num_samples

        # set the MCMC kwargs.
        if mcmc_kwargs is not None:
            self.mcmc_kwargs = mcmc_kwargs
        else:
            self.mcmc_kwargs = {}
        if nuts_kwargs is not None:
            self.nuts_kwargs = nuts_kwargs
        else:
            self.nuts_kwargs = {}

    def _reduce(self, x, y):
        # set up a sparse regression and sample.
        self.mcmc_ = self._run_mcmc(x, y, self.nuts_kwargs, self.mcmc_kwargs)

        # set the mean values as the coefficients.
        self.coef_ = np.array(self.mcmc_.get_samples()["beta"].mean(axis=0))

    def _numpyro_model(self, x, y):
        # get the dimensionality of the problem.
        n_features = x.shape[1]
        n_targets = y.shape[1]

        # sample the horseshoe hyperparameters.
        tau = numpyro.sample("tau", HalfCauchy(self.sparsity_coef_tau0))
        c_sq = numpyro.sample(
            "c_sq",
            InverseGamma(
                self.slab_shape_nu / 2, self.slab_shape_nu / 2 * self.slab_shape_s**2
            ),
        )

        # sample the parameters compute the predicted outputs.
        beta = _sample_reg_horseshoe(tau, c_sq, (n_targets, n_features))
        mu = jnp.dot(x, beta.T)

        # compute the likelihood.
        sigma = numpyro.sample("sigma", Exponential(self.noise_hyper_lambda))
        numpyro.sample("obs", Normal(mu, sigma), obs=y)

    def _run_mcmc(self, x, y, nuts_kwargs, mcmc_kwargs):
        # set up a jax random key.
        seed = mcmc_kwargs.pop("seed", 0)
        rng_key = random.PRNGKey(seed)

        # run the MCMC
        kernel = NUTS(self._numpyro_model, **nuts_kwargs)
        mcmc = MCMC(
            kernel,
            num_warmup=self.num_warmup,
            num_samples=self.num_samples,
            **mcmc_kwargs,
        )
        mcmc.run(rng_key, x=x, y=y)

        # extract the MCMC samples and compute the UQ-SINDy parameters.
        return mcmc


def _sample_reg_horseshoe(tau: float, c_sq: float, shape: Tuple[int, ...]):
    """Create a regularized horseshoe distribution

    The regularized horseshoe distribution behaves like a horseshoe prior when
    shrinkage ``lamb`` is small, but behaves like a gaussian slab of variance
    ``c_sq`` when ``lamb`` is big or a Student T-slab when ``c_sq`` is itself
    an inverse Gamma.

    For original work, including interpretation of the coefficients, see:

    Piironen, J., and Vehtari, A. (2017). Sparsity Information and
    Regularization in the Horseshoe and Other Shrinkage Priors. Electronic Journal
    of Statistics Vol. 11 pp 5018-5051. https://doi.org/10.1214/17-EJS1337SI
    """
    lamb = numpyro.sample("lambda", HalfCauchy(1.0), sample_shape=shape)
    lamb_squiggle = jnp.sqrt(c_sq) * lamb / jnp.sqrt(c_sq + tau**2 * lamb**2)
    beta = numpyro.sample(
        "beta",
        Normal(jnp.zeros_like(lamb_squiggle), jnp.sqrt(lamb_squiggle**2 * tau**2)),
    )
    return beta



import os
import runpy
import sys
from contextlib import contextmanager
from pathlib import Path

import nbformat
import pytest
from ipykernel import kernelapp as app
from nbconvert.preprocessors import ExecutePreprocessor

# Find all example notebooks
notebook_dir = Path(__file__).parent.parent / "examples"
notebooks = [nb.name for nb in notebook_dir.glob("*.ipynb")]
notebook_scripts = []
for x in notebook_dir.iterdir():
    if x.is_dir() and (x / "example.py").exists():
        notebook_scripts.append(x.name)


def _load_notebook(filename: Path) -> nbformat.NotebookNode:
    """Load Jupyter Notebook as NotebookNode object."""
    with open(filename, "r") as fh:
        notebook = nbformat.read(fh, as_version=4)
    return notebook


def _create_kernel() -> str:
    """Create a IPython Kernel for the current executable."""
    kernel_name = sys.executable.replace("/", ".").replace("\\", ".").replace(":", ".")
    app.launch_new_instance(argv=["install", "--user", "--name", kernel_name])
    return kernel_name


@contextmanager
def _cwd(directory):
    """Modify working directory and sys.path temporarily"""
    cwd = Path.cwd()
    os.chdir(directory)
    sys.path.insert(0, str(directory))
    yield
    sys.path.pop(0)
    os.chdir(cwd)


@pytest.fixture
def purge_notebook_modules():
    """Remove notebook modules from sys.modules after test.

    Because these modules share common names in each notebook and
    module names have a system-wide scope, import machinery will not
    import new modules for successive notebooks unless old modules of
    same name are removed from sys.modules.

    This might be better served by fixing imports in notebooks using
    importlib.
    """
    local_modules = [
        key for key in sys.modules if key[:5] == "utils" or key[:9] == "mock_data"
    ]
    [sys.modules.pop(mod) for mod in local_modules]
    yield
    SENTINEL = object()
    [sys.modules.pop(mod, SENTINEL) for mod in local_modules]


@pytest.mark.parametrize("directory", notebook_scripts)
@pytest.mark.notebooks
def test_notebook_script(directory: Path, purge_notebook_modules):
    # Run in native directory with modified sys.path for imports to work
    if "17_" in directory or "5_" in directory or "8_" in directory:
        pytest.skip("Notebook runs too slowly.  Test manually")
    with _cwd(notebook_dir / directory):
        try:
            runpy.run_path(
                str(notebook_dir / directory / "example.py"), run_name="testing"
            )
        except SystemExit:
            pass


def test_external(external_notebook: Path, purge_notebook_modules):
    with _cwd(external_notebook.resolve()):
        try:
            runpy.run_path(str("example.py"), run_name="testing")
        except SystemExit:
            pass


@pytest.mark.parametrize("filename", notebooks)
@pytest.mark.slow
def test_notebook(filename):
    nb = _load_notebook(notebook_dir / filename)
    # Run in native directory to find data via relative path
    with _cwd(notebook_dir):
        ExecutePreprocessor(timeout=-1, kernel=_create_kernel).preprocess(nb)



import numpy as np
import numpy.typing as npt

# In python 3.12, use type statement
# https://docs.python.org/3/reference/simple_stmts.html#the-type-statement
NpFlt = np.floating[npt.NBitBase]
FloatDType = np.dtype[np.floating[npt.NBitBase]]
Int1D = np.ndarray[tuple[int], np.dtype[np.int_]]
Float1D = np.ndarray[tuple[int], FloatDType]
Float2D = np.ndarray[tuple[int, int], FloatDType]
Float3D = np.ndarray[tuple[int, int, int], FloatDType]
Float4D = np.ndarray[tuple[int, int, int, int], FloatDType]
Float5D = np.ndarray[tuple[int, int, int, int, int], FloatDType]
FloatND = npt.NDArray[NpFlt]



import warnings
from copy import deepcopy
from typing import Optional
from typing import Tuple

try:
    import cvxpy as cp

    cvxpy_flag = True
except ImportError:
    cvxpy_flag = False
    pass
import numpy as np
from scipy.linalg import cho_factor
from sklearn.exceptions import ConvergenceWarning

from ..utils import reorder_constraints
from .sr3 import SR3


class ConstrainedSR3(SR3):
    """
    Sparse relaxed regularized regression with linear (in)equality constraints.

    Attempts to minimize the objective function

    .. math::

        0.5\\|y-Xw\\|^2_2 + \\lambda R(u)
        + (0.5 / \\nu)\\|w-u\\|^2_2

    .. math::

        \\text{subject to } Cw = d

    over u and w, where :math:`R(u)` is a regularization function, C is a
    constraint matrix, and d is a vector of values. See the following
    reference for more details:

        Champion, Kathleen, et al. "A unified sparse optimization framework
        to learn parsimonious physics-informed models from data."
        IEEE Access 8 (2020): 169259-169271.

        Zheng, Peng, et al. "A unified framework for sparse relaxed
        regularized regression: Sr3." IEEE Access 7 (2018): 1404-1423.

    Parameters
    ----------
    constraint_lhs : numpy ndarray, optional (default None)
        Shape should be (n_constraints, n_features * n_targets),
        The left hand side matrix C of Cw <= d (Or Cw = d for equality
        constraints). There should be one row per constraint.

    constraint_rhs : numpy ndarray, shape (n_constraints,), optional (default None)
        The right hand side vector d of Cw <= d.

    constraint_order : string, optional (default "target")
        The format in which the constraints ``constraint_lhs`` were passed.
        Must be one of "target" or "feature".
        "target" indicates that the constraints are grouped by target:
        i.e. the first ``n_features`` columns
        correspond to constraint coefficients on the library features
        for the first target (variable), the next ``n_features`` columns to
        the library features for the second target (variable), and so on.
        "feature" indicates that the constraints are grouped by library
        feature: the first ``n_targets`` columns correspond to the first
        library feature, the next ``n_targets`` columns to the second library
        feature, and so on.

    inequality_constraints : bool, optional (default False)
        If True, CVXPY methods are used to solve the problem.

    verbose_cvxpy : bool, optional (default False)
        Boolean flag which is passed to CVXPY solve function to indicate if
        output should be verbose or not. Only relevant for optimizers that
        use the CVXPY package in some capabity.

    See base class for additional arguments

    Attributes
    ----------
    coef_ : array, shape (n_features,) or (n_targets, n_features)
        Regularized weight vector(s). This is the v in the objective
        function.

    coef_full_ : array, shape (n_features,) or (n_targets, n_features)
        Weight vector(s) that are not subjected to the regularization.
        This is the w in the objective function.

    history_ : list
        History of sparse coefficients. ``history_[k]`` contains the
        sparse coefficients (v in the optimization objective function)
        at iteration k.

    objective_history_ : list
        History of the value of the objective at each step. Note that
        the trapping SINDy problem is nonconvex, meaning that this value
        may increase and decrease as the algorithm works.
    """

    def __init__(
        self,
        reg_weight_lam=0.005,
        regularizer="l0",
        relax_coeff_nu=1.0,
        tol=1e-5,
        max_iter=30,
        trimming_fraction=0.0,
        trimming_step_size=1.0,
        constraint_lhs=None,
        constraint_rhs=None,
        constraint_order="target",
        normalize_columns=False,
        copy_X=True,
        initial_guess=None,
        equality_constraints=False,
        inequality_constraints=False,
        constraint_separation_index: Optional[bool] = None,
        verbose=False,
        verbose_cvxpy=False,
        unbias=False,
    ):
        super().__init__(
            reg_weight_lam=reg_weight_lam,
            regularizer=regularizer,
            relax_coeff_nu=relax_coeff_nu,
            tol=tol,
            trimming_fraction=trimming_fraction,
            trimming_step_size=trimming_step_size,
            max_iter=max_iter,
            initial_guess=initial_guess,
            copy_X=copy_X,
            normalize_columns=normalize_columns,
            verbose=verbose,
            unbias=unbias,
        )

        self.verbose_cvxpy = verbose_cvxpy
        self.constraint_lhs = constraint_lhs
        self.constraint_rhs = constraint_rhs
        self.constraint_order = constraint_order
        self.use_constraints = (constraint_lhs is not None) or (
            constraint_rhs is not None
        )

        if (
            self.use_constraints
            and not equality_constraints
            and not inequality_constraints
        ):
            warnings.warn(
                "constraint_lhs and constraint_rhs passed to the optimizer, "
                " but user did not specify if the constraints were equality or"
                " inequality constraints. Assuming equality constraints."
            )
            equality_constraints = True

        if self.use_constraints:
            if constraint_order not in ("feature", "target"):
                raise ValueError(
                    "constraint_order must be either 'feature' or 'target'"
                )
            if unbias:
                raise ValueError(
                    "Constraints are incompatible with an unbiasing step.  Set"
                    " unbias=False"
                )

        if inequality_constraints and not cvxpy_flag:
            raise ValueError(
                "Cannot use inequality constraints without cvxpy installed."
            )

        if inequality_constraints and not self.use_constraints:
            raise ValueError(
                "Use of inequality constraints requires constraint_lhs and "
                "constraint_rhs."
            )

        if inequality_constraints and regularizer.lower() not in (
            "l1",
            "l2",
            "weighted_l1",
            "weighted_l2",
        ):
            raise ValueError(
                "Use of inequality constraints requires a convex regularizer."
            )
        self.inequality_constraints = inequality_constraints
        self.equality_constraints = equality_constraints
        if self.use_constraints and constraint_separation_index is None:
            if self.inequality_constraints and not self.equality_constraints:
                constraint_separation_index = len(constraint_lhs)
            elif self.equality_constraints and not self.inequality_constraints:
                constraint_separation_index = 0
            else:
                raise ValueError(
                    "If passing both inequality and equality constraints, must specify"
                    " constraint_separation_index."
                )
        self.constraint_separation_index = constraint_separation_index

    def _update_full_coef_constraints(self, H, x_transpose_y, coef_sparse):
        g = x_transpose_y + coef_sparse / self.relax_coeff_nu
        inv1 = np.linalg.inv(H)
        inv1_mod = np.kron(inv1, np.eye(coef_sparse.shape[1]))
        inv2 = np.linalg.inv(
            self.constraint_lhs.dot(inv1_mod).dot(self.constraint_lhs.T)
        )

        rhs = g.flatten() + self.constraint_lhs.T.dot(inv2).dot(
            self.constraint_rhs - self.constraint_lhs.dot(inv1_mod).dot(g.flatten())
        )
        rhs = rhs.reshape(g.shape)
        return inv1.dot(rhs)

    @staticmethod
    def _calculate_penalty(
        regularization: str, regularization_weight, xi: cp.Variable
    ) -> cp.Expression:
        """
        Args:
        -----
        regularization: 'l0' | 'weighted_l0' | 'l1' | 'weighted_l1' |
                        'l2' | 'weighted_l2'
        regularization_weight: float | np.array, can be a scalar
                               or an array of the same shape as xi
        xi: cp.Variable

        Returns:
        --------
        cp.Expression
        """
        regularization = regularization.lower()
        if regularization == "l1":
            return regularization_weight * cp.sum(cp.abs(xi))
        elif regularization == "weighted_l1":
            return cp.sum(cp.multiply(regularization_weight, cp.abs(xi)))
        elif regularization == "l2":
            return regularization_weight * cp.sum(xi**2)
        elif regularization == "weighted_l2":
            return cp.sum(cp.multiply(regularization_weight, xi**2))

    def _create_var_and_part_cost(
        self, var_len: int, x_expanded: np.ndarray, y: np.ndarray
    ) -> Tuple[cp.Variable, cp.Expression]:
        xi = cp.Variable(var_len)
        cost = cp.sum_squares(x_expanded @ xi - y.flatten())
        penalty = self._calculate_penalty(
            self.regularizer, np.ravel(self.reg_weight_lam), xi
        )
        return xi, cost + penalty

    def _update_coef_cvxpy(self, xi, cost, var_len, coef_prev, tol):
        if self.use_constraints:
            constraints = []
            if self.equality_constraints:
                constraints.append(
                    self.constraint_lhs[self.constraint_separation_index :, :] @ xi
                    == self.constraint_rhs[self.constraint_separation_index :],
                )
            if self.inequality_constraints:
                constraints.append(
                    self.constraint_lhs[: self.constraint_separation_index, :] @ xi
                    <= self.constraint_rhs[: self.constraint_separation_index]
                )
            prob = cp.Problem(cp.Minimize(cost), constraints)
        else:
            prob = cp.Problem(cp.Minimize(cost))

        prob_clone = deepcopy(prob)
        try:
            prob.solve(
                max_iter=self.max_iter,
                eps_abs=tol,
                eps_rel=tol,
                verbose=self.verbose_cvxpy,
            )
        except cp.error.SolverError:
            try:
                prob = prob_clone
                prob.solve(max_iter=self.max_iter, verbose=self.verbose_cvxpy)
                xi = prob.variables()[0]
            except cp.error.SolverError:
                warnings.warn("Solver failed, setting coefs to zeros")
                xi.value = np.zeros(var_len)

        if xi.value is None:
            warnings.warn(
                "Infeasible solve, probably an issue with the regularizer "
                " or the constraint that was used.",
                ConvergenceWarning,
            )
            return None
        coef_new = (xi.value).reshape(coef_prev.shape)
        return coef_new

    def _reduce(self, x, y):
        """
        Perform at most ``self.max_iter`` iterations of the SR3 algorithm
        with inequality constraints.

        Assumes initial guess for coefficients is stored in ``self.coef_``.
        """
        if self.initial_guess is not None:
            self.coef_ = self.initial_guess

        coef_sparse = self.coef_.T
        coef_full = coef_sparse.copy()
        n_samples, n_features = x.shape
        n_targets = y.shape[1]

        if self.use_trimming:
            trimming_array = np.repeat(1.0 - self.trimming_fraction, n_samples)
            self.history_trimming_ = [trimming_array]

        if self.use_constraints and self.constraint_order.lower() == "target":
            self.constraint_lhs = reorder_constraints(self.constraint_lhs, n_features)

        # Precompute some objects for upcoming least-squares solves.
        # Assumes that self.relax_coeff_nu is fixed throughout optimization procedure.
        H = np.dot(x.T, x) + np.diag(np.full(x.shape[1], 1.0 / self.relax_coeff_nu))
        x_transpose_y = np.dot(x.T, y)
        if not self.use_constraints:
            cho = cho_factor(H)
        if self.inequality_constraints:
            # Precompute some objects for optimization
            x_expanded = np.zeros((n_samples, n_targets, n_features, n_targets))
            for i in range(n_targets):
                x_expanded[:, i, :, i] = x
            x_expanded = np.reshape(
                x_expanded, (n_samples * n_targets, n_targets * n_features)
            )

        # Print initial values for each term in the optimization
        if self.verbose:
            row = [
                "Iteration",
                "|y - Xw|^2",
                "|w-u|^2/v",
                "R(u)",
                "Total Error: |y - Xw|^2 + |w - u|^2 / v + R(u)",
            ]
            print(
                "{: >10} ... {: >10} ... {: >10} ... {: >10} ... {: >10}".format(*row)
            )

        objective_history = []
        if self.inequality_constraints:
            var_len = coef_sparse.shape[0] * coef_sparse.shape[1]
            xi, cost = self._create_var_and_part_cost(var_len, x_expanded, y)
            coef_sparse = self._update_coef_cvxpy(
                xi, cost, var_len, coef_sparse, self.tol
            )
            objective_history.append(self._objective(x, y, 0, coef_full, coef_sparse))
        else:
            for k in range(self.max_iter):
                if self.use_trimming:
                    x_weighted = x * trimming_array.reshape(n_samples, 1)
                    H = np.dot(x_weighted.T, x) + np.diag(
                        np.full(x.shape[1], 1.0 / self.relax_coeff_nu)
                    )
                    x_transpose_y = np.dot(x_weighted.T, y)
                    if not self.use_constraints:
                        cho = cho_factor(H)
                    trimming_grad = 0.5 * np.sum((y - x.dot(coef_full)) ** 2, axis=1)
                if self.use_constraints:
                    coef_full = self._update_full_coef_constraints(
                        H, x_transpose_y, coef_sparse
                    )
                else:
                    coef_full = self._update_full_coef(cho, x_transpose_y, coef_sparse)
                coef_sparse = self._update_sparse_coef(coef_full)
                self.history_.append(np.copy(coef_sparse).T)

                if self.use_trimming:
                    trimming_array = self._update_trimming_array(
                        coef_full, trimming_array, trimming_grad
                    )

                    objective_history.append(
                        self._objective(x, y, k, coef_full, coef_sparse, trimming_array)
                    )
                else:
                    objective_history.append(
                        self._objective(x, y, k, coef_full, coef_sparse)
                    )
                if self._convergence_criterion() < self.tol:
                    # TODO: Update this for trimming/constraints
                    break
            else:
                warnings.warn(
                    f"ConstrainedSR3 did not converge after {self.max_iter}"
                    " iterations.",
                    ConvergenceWarning,
                )
        if self.use_constraints and self.constraint_order.lower() == "target":
            self.constraint_lhs = reorder_constraints(
                self.constraint_lhs, n_features, output_order="target"
            )
        self.coef_ = coef_sparse.T
        self.coef_full_ = coef_full.T
        if self.use_trimming:
            self.trimming_array = trimming_array
        self.objective_history = objective_history



from typing import List
from typing import Union

import numpy as np
from numpy.typing import NDArray
from scipy.special import factorial

from .base import BaseDifferentiation
from pysindy.utils._axes import AxesArray


class FiniteDifference(BaseDifferentiation):
    """Finite difference derivatives.

    Only centered differences are implemented, for even order and
    left-off-centered differences for odd order. See
    `wikipedia <https://en.wikipedia.org/wiki/Finite_difference_coefficient>`_
    for the coefficient formula used.

    Parameters
    ----------
    order: int, optional (default 2)
        The accuracy big-O exponent of the finite difference method to be used.

    d : int, optional (default 1)
        The order of derivative to take.  Must be positive integer.

    axis: int, optional (default 0)
        The axis to differentiate along.

    is_uniform : boolean, optional (default False)
        Parameter to tell the differentiation that, although a N-dim
        grid is passed, it is uniform so can use dx instead of the full
        grid array.

    drop_endpoints: boolean, optional (default False)
        Whether or not derivatives are computed for endpoints.
        If False, endpoints will be set to np.nan.
        Note that which points are endpoints depends on the method
        being used.

    periodic: boolean, optional (default False)
        Whether to use periodic boundary conditions for endpoints.
        Use forward differences for periodic=False and periodic boundaries
        with centered differences for periodic=True on the boundaries.
        No effect if drop_endpoints=True

    Attributes:
        n_stencil: the number of coefficients in the finite difference expression.

    Examples
    --------
    >>> import numpy as np
    >>> from pysindy.differentiation import FiniteDifference
    >>> t = np.linspace(0, 1, 5)
    >>> X = np.vstack((np.sin(t), np.cos(t))).T
    >>> fd = FiniteDifference()
    >>> fd._differentiate(X, t)
    array([[ 1.00114596,  0.00370551],
           [ 0.95885108, -0.24483488],
           [ 0.8684696 , -0.47444711],
           [ 0.72409089, -0.67456051],
           [ 0.53780339, -0.84443737]])
    """

    def __init__(
        self,
        order: int = 2,
        d: int = 1,
        axis: int = 0,
        is_uniform: bool = False,
        drop_endpoints: bool = False,
        periodic: bool = False,
    ):
        if order <= 0 or not isinstance(order, int):
            raise ValueError("order must be a positive int")
        if d <= 0:
            raise ValueError("differentiation order must be a positive int")

        self.d = d
        self.order = order
        self.is_uniform = is_uniform
        self.axis = axis
        self.drop_endpoints = drop_endpoints
        self.periodic = periodic
        self.n_stencil = int(2 * ((self.d + 1) // 2) - 1 + self.order)
        self.n_stencil_forward = self.d + self.order

        if self.d >= self.n_stencil:
            raise ValueError(
                "This combination of d and order is not implemented. "
                "It is required that d >= stencil_size, where "
                "stencil_size = 2 * (d + 1) // 2 - 1 + order. "
            )

    def _coefficients(self, t):
        nt = len(t)
        self.stencil_inds = AxesArray(
            np.array(
                [
                    np.arange(i, nt - self.n_stencil + i + 1)
                    for i in range(self.n_stencil)
                ]
            ),
            {"ax_offset": 0, "ax_ti": 1},
        )
        self.stencil = AxesArray(
            np.transpose(t[self.stencil_inds]), {"ax_time": 0, "ax_offset": 1}
        )
        pows = np.arange(self.n_stencil)[np.newaxis, :, np.newaxis]
        dt_endpoints = (
            self.stencil
            - t[(self.n_stencil - 1) // 2 : -(self.n_stencil - 1) // 2, "offset"]
        )
        matrices = dt_endpoints[:, "power", :] ** pows
        b = AxesArray(np.zeros((self.n_stencil)), {"ax_power": 0})
        b[self.d] = factorial(self.d)
        return np.linalg.solve(matrices, b)

    def _coefficients_boundary_forward(self, t):
        # use the same stencil for each boundary point,
        # but change the evaluation point
        left = np.arange(self.n_stencil_forward)[:, np.newaxis] * np.ones(
            (self.n_stencil - 1) // 2, dtype=int
        )
        if self.order % 2 == 0:
            right_len = (self.n_stencil - 1) // 2
        else:
            right_len = 1 + (self.n_stencil - 1) // 2
        right = (-1 - np.arange(self.n_stencil_forward))[:, np.newaxis] * np.ones(
            right_len, dtype=int
        )
        tinds = np.concatenate(
            [
                np.arange((self.n_stencil - 1) // 2, dtype=int),
                np.flip(-1 - np.arange(right_len, dtype=int)),
            ]
        )
        self.stencil_inds = np.concatenate([left, right], axis=1)

        pows = np.arange(self.n_stencil_forward)[np.newaxis, :, np.newaxis]

        if np.isscalar(t):
            matrices = np.transpose(
                (t * (self.stencil_inds - tinds)[:, np.newaxis, :]) ** pows
            )
        else:
            matrices = np.transpose(
                ((t[self.stencil_inds] - t[tinds])[:, np.newaxis, :]) ** pows
            )

        b = np.zeros((self.n_stencil_forward,))
        b[self.d] = factorial(self.d)
        return np.linalg.solve(matrices, b)

    def _coefficients_boundary_periodic(self, t):
        # use centered periodic stencils
        left = (np.arange(self.n_stencil) - (self.n_stencil - 1) // 2)[
            :, np.newaxis
        ] + np.arange((self.n_stencil - 1) // 2, dtype=int)
        right = np.flip(
            (-1 - np.arange(self.n_stencil) + (self.n_stencil - 1) // 2)[:, np.newaxis]
            - np.arange((self.n_stencil - 1) // 2, dtype=int),
            axis=1,
        )
        self.stencil_inds = np.concatenate([left, right], axis=1)
        tinds = np.concatenate(
            [
                np.arange((self.n_stencil - 1) // 2, dtype=int),
                np.flip(-1 - np.arange((self.n_stencil - 1) // 2, dtype=int)),
            ]
        )
        pows = np.arange(self.n_stencil)[np.newaxis, :, np.newaxis]

        if np.isscalar(t):
            matrices = (
                np.transpose(
                    t
                    * (
                        np.concatenate(
                            [
                                np.ones((self.n_stencil - 1) // 2),
                                -np.ones((self.n_stencil - 1) // 2),
                            ]
                        )
                        * (np.arange(self.n_stencil) - (self.n_stencil - 1) // 2)[
                            :, np.newaxis
                        ]
                    )[:, np.newaxis, :]
                )
                ** pows
            )
        else:
            period = t[-1] - t[0] + (t[1] - t[0])
            matrices = np.transpose(
                (
                    (
                        np.mod(t[self.stencil_inds] - t[tinds] + period / 2, period)
                        - period / 2
                    )[:, np.newaxis, :]
                )
                ** pows
            )

        b = np.zeros(self.n_stencil)
        b[self.d] = factorial(self.d)
        return np.linalg.solve(matrices, b)

    def _constant_coefficients(self, dt):
        pows = np.arange(self.n_stencil)[:, np.newaxis]
        matrices = (dt * (np.arange(self.n_stencil) - (self.n_stencil - 1) // 2))[
            np.newaxis, :
        ] ** pows
        b = np.zeros(self.n_stencil)
        b[self.d] = factorial(self.d)
        return np.linalg.solve(matrices, b)

    def _accumulate(self, coeffs, x):
        # slice to select the stencil indices
        s = [slice(None)] * x.ndim
        s[self.axis] = self.stencil_inds

        # a new axis is introduced before self.axis for the stencil indices
        # To contract with the coefficients, roll by -self.axis to put it first
        # Then roll back by self.axis to return the order
        trans = np.roll(np.arange(x.ndim + 1), -self.axis)
        # TODO: assign x's axes much earlier in the call stack
        x = AxesArray(x, {"ax_unk": list(range(x.ndim))})
        x_expanded = AxesArray(
            np.transpose(x[tuple(s)], axes=trans), x.insert_axis(0, "ax_offset")
        )
        return np.transpose(
            np.einsum(
                "ij...,ij->j...",
                x_expanded,
                np.transpose(coeffs),
            ),
            np.roll(np.arange(x.ndim), self.axis),
        )

    def _differentiate(
        self, x: NDArray, t: Union[NDArray, float, List[float]]
    ) -> NDArray:
        """
        Apply finite difference method.
        """
        x_dot = np.full_like(x, fill_value=np.nan)
        s = [slice(None)] * len(x.shape)

        if self.axis < 0:
            # Need to do this for _accumulate function to work properly?
            self.axis = len(x.shape) + self.axis

        # Central differences in interior of domain
        if np.isscalar(t) or self.is_uniform:
            dt = t
            if not np.isscalar(t):
                dt = t[1] - t[0]

            coeffs = self._constant_coefficients(dt)
            dims = np.array(x.shape)
            dims[self.axis] = x.shape[self.axis] - (self.n_stencil - 1)
            interior = np.zeros(dims)
            # Slightly faster version of self._accumulate for uniform grid
            for i in range(self.n_stencil):
                if abs(coeffs[i]) > 0:
                    start = i
                    stop = -(self.n_stencil - start - 1)
                    if stop >= 0:
                        stop = None
                    s[self.axis] = slice(start, stop)
                    interior = interior + x[tuple(s)] * coeffs[i]
        else:
            t = AxesArray(np.array(t), axes={"ax_time": 0})
            coeffs = self._coefficients(t)
            interior = self._accumulate(coeffs, x)
        s[self.axis] = slice((self.n_stencil - 1) // 2, -(self.n_stencil - 1) // 2)
        x_dot[tuple(s)] = interior

        # Boundaries
        if not self.drop_endpoints:
            # Forward differences on boundary
            if not self.periodic:
                coeffs = self._coefficients_boundary_forward(t)
                boundary = self._accumulate(coeffs, x)

                if self.order % 2 == 0:
                    right_len = (self.n_stencil - 1) // 2
                else:
                    right_len = 1 + (self.n_stencil - 1) // 2
                s[self.axis] = np.concatenate(
                    [
                        np.arange((self.n_stencil - 1) // 2, dtype=int),
                        np.flip(-1 - np.arange(right_len, dtype=int)),
                    ]
                )
            # Central differences on boundary with periodic bcs
            else:
                coeffs = self._coefficients_boundary_periodic(t)
                boundary = self._accumulate(coeffs, x)
                s[self.axis] = np.concatenate(
                    [
                        np.arange(0, (self.n_stencil - 1) // 2),
                        -np.flip(1 + np.arange(1, (self.n_stencil - 1) // 2)),
                        np.array([-1]),
                    ]
                )
            x_dot[tuple(s)] = boundary
        self.smoothed_x_ = x
        return x_dot



import numpy as np

from pysindy import FiniteDifference


class FiniteDifferenceBM:
    """
    An example benchmark that times the performance of various kinds
    of iterating over dictionaries in Python.
    """

    def setup(self):
        self.differentiator = FiniteDifference()
        self.t = np.arange(0, 2 * np.pi, 0.01)

    def time_derivative(self):
        self.differentiator(self.t, t=self.t)

    def peakmem_derivative(self):
        self.differentiator(self.t, t=self.t)



import importlib
import os
import re
import shutil
from pathlib import Path
from typing import TypeVar

import requests
import yaml
from docutils import nodes
from docutils.statemachine import StringList
from sphinx.application import Sphinx
from sphinx.directives.other import TocTree
from sphinx.util.docutils import SphinxDirective

author = "dynamicslab"
project = "pysindy"
copyright = f"2020, {author}"
module = importlib.import_module(project)
version = release = getattr(module, "__version__")

master_doc = "index"

extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.autosummary",
    "nbsphinx",
    "sphinx.ext.todo",
    "sphinx.ext.napoleon",
    "sphinx.ext.viewcode",
    "sphinx.ext.mathjax",
    "sphinx.ext.intersphinx",
    "IPython.sphinxext.ipython_console_highlighting",
    "matplotlib.sphinxext.plot_directive",
]

nb_execution_mode = "off"

templates_path = ["templates"]
autosummary_generate = True
autosummary_ignore_module_all = False

autoclass_content = "both"

language = "en"

here = Path(__file__).parent.resolve()

exclude_patterns = ["build", "_build", "Youtube"]
# pygments_style = "sphinx"

add_module_names = True
add_function_parentheses = False
todo_include_todos = True

html_theme = "sphinx_rtd_theme"
html_show_sourcelink = False
html_show_sphinx = False
html_show_copyright = True

plot_html_show_source_link = False
plot_html_show_formats = False

default_role = "any"
html_sourcelink_suffix = ""

intersphinx_mapping = {
    "derivative": ("https://derivative.readthedocs.io/en/latest", None),
    "sklearn": ("https://scikit-learn.org/stable", None),
    "numpy": ("https://numpy.org/doc/stable", None),
}

show_warning_types = True
suppress_warnings = [
    "ref.python",
    "ref.exc",
    "ref.class",
    "ref.obj",
    "ref.meth",
    "ref.any",
    "ref",
]

# -- Extensions to the  Napoleon GoogleDocstring class ---------------------
# michaelgoerz.net/notes/extending-sphinx-napoleon-docstring-sections.html
from sphinx.ext.napoleon.docstring import GoogleDocstring  # noqa: E402


def parse_keys_section(self, section):
    return self._format_fields("Keys", self._consume_fields())


GoogleDocstring._parse_keys_section = parse_keys_section


def parse_attributes_section(self, section):
    return self._format_fields("Attributes", self._consume_fields())


GoogleDocstring._parse_attributes_section = parse_attributes_section


def parse_class_attributes_section(self, section):
    return self._format_fields("Class Attributes", self._consume_fields())


GoogleDocstring._parse_class_attributes_section = parse_class_attributes_section


def patched_parse(self):
    """
    we now patch the parse method to guarantee that the the above methods are
    assigned to the _section dict
    """
    self._sections["keys"] = self._parse_keys_section
    self._sections["class attributes"] = self._parse_class_attributes_section
    self._unpatched_parse()


GoogleDocstring._unpatched_parse = GoogleDocstring._parse
GoogleDocstring._parse = patched_parse


def setup(app: Sphinx):
    """Our sphinx extension for copying from examples/ to docs/examples

    Since nbsphinx does not handle glob/regex paths, we need to
    manually copy documentation source files from examples.  See issue
    # 230.
    """
    doc_examples = here / "examples"
    if not doc_examples.exists():
        (here / "examples").mkdir()
    example_source = (here / "../examples").resolve()
    source_notebooks = example_source.glob("**/*.ipynb")
    shutil.copy(example_source / "README.rst", doc_examples / "index.rst")
    for notebook in source_notebooks:
        if notebook.parent == example_source:
            new_dir = doc_examples / notebook.stem
        else:
            new_dir = doc_examples / notebook.parent.stem
        new_dir.mkdir(exist_ok=True)
        new_file = new_dir / "example.ipynb"
        print(f"Copying {notebook} to {new_file}")
        shutil.copy(notebook, new_file)
    # Notebook 15 uses an image file
    (doc_examples / "15_pysindy_lectures/data").mkdir(exist_ok=True)
    shutil.copy(
        example_source / "data/optimizer_summary.jpg",
        doc_examples / "15_pysindy_lectures/data/optimizer_summary.jpg",
    )
    if (here / "static/custom.css").exists():
        app.add_css_file("custom.css")

    _grab_external_examples(example_source)
    app.add_directive("pysindy-example", PysindyExample)


EXTERNAL_EXAMPLES: dict[str, list[tuple[str, str]]] = {}


def _load_ext_config(example_source: Path) -> list[dict[str, str]]:
    ext_config = example_source / "external.yml"
    with open(ext_config) as f:
        ext_examples = yaml.safe_load(f)
    return ext_examples


def _grab_external_examples(example_source: Path):
    ext_examples = _load_ext_config(example_source)
    for example in ext_examples:
        ex_name = example["name"]
        user = example["user"]
        repo = example["repo"]
        ref = example["ref"]
        dir = example["dir"]
        base = f"https://raw.githubusercontent.com/{user}/{repo}/{ref}/{dir}/"
        notebooks = fetch_notebook_list(base)
        base = f"https://raw.githubusercontent.com/{user}/{repo}/{ref}/"
        local_nbs = [(name, _copy_nb(base, pth, repo, ref)) for name, pth in notebooks]
        EXTERNAL_EXAMPLES[ex_name] = local_nbs


class PysindyExample(SphinxDirective):
    """Create an entry in a list of examples.

    The directive requires a title and key, which is a reference to external.yml.
    See _load_ext_config and ../examples/external.yml for the required linking.

    The directive attempts to create a level-two header, unless the directive
    appears before any other level-1 header.
    """

    required_arguments = 0
    optional_arguments = 0
    option_spec = {"key": str, "title": str}
    has_content = True

    def run(self) -> list[nodes.Node]:
        key = self.options["key"]
        example_config = _load_ext_config((here / "../examples").resolve())
        try:
            this_example = [ex for ex in example_config if ex["name"] == key][0]
        except IndexError:
            RuntimeError("Unknown configuration key for external example")
        heading_text: str = self.options.get("title")
        base_repo = f"https://github.com/{this_example['user']}/{this_example['repo']}"
        repo_ref = nodes.reference(
            name="Source repo",
            text=" Source repo",
            refuri=f"{base_repo}/tree/{this_example['ref']}",
        )
        normalized_text = re.sub(r"\s", "_", heading_text)
        tgt_node = nodes.target(refid=normalized_text)
        title_node = nodes.title()
        title_text = self.parse_inline(heading_text)[0]
        title_node += [*title_text, tgt_node]
        content_nodes = self.parse_content_to_nodes()
        if isinstance(content_nodes[-1], nodes.paragraph):
            repo_par = content_nodes[-1]
        else:
            repo_par = nodes.paragraph()
            content_nodes += [repo_par]
        repo_par += repo_ref
        toc_items = []
        for name, relpath in EXTERNAL_EXAMPLES[key]:
            if name:
                toc_str = f"{name} <{relpath}>"
            if not name:
                toc_str = relpath
            toc_items.append(toc_str)
        toc_nodes = TocTree(
            name="PysindyExample",
            options={"maxdepth": 1, "titlesonly": True},
            arguments=[],
            content=StringList(initlist=toc_items),
            lineno=self.lineno,
            block_text="",
            content_offset=self.content_offset,
            state=self.state,
            state_machine=self.state_machine,
        ).run()
        section_node = nodes.section(ids=[heading_text], names=[heading_text])
        section_node += [title_node, *content_nodes, *toc_nodes]
        # We need to find section level 1, and be it's child.
        parent = self.state.parent
        # If section level 1 or 2 haven't been created yet, just insert in position
        if not isinstance(parent, nodes.section):
            return [section_node]
        # Otherwise, crawl up document tree until next_parent is document itself.
        while isinstance(next_parent := parent.parent, nodes.section):
            parent = next_parent
            next_parent = next_parent.parent
        parent += section_node
        return []


def fetch_notebook_list(base: str) -> list[tuple[str, str]]:
    """Gets the list of example notebooks from a repo's index.html

    Each entry is a tuple of the title name of a link and the address
    """
    index = requests.get(base + "index.rst")
    if index.status_code != 200:
        raise RuntimeError(f"Unable to locate external example directory for {base}")
    text = str(index.content, encoding="utf-8")
    link_line = r"^\s+(.*)[^\S\r\n]+(\S+.ipynb)"
    T = TypeVar("T")

    def deduplicate(mylist: list[T]) -> list[T]:
        return list(set(mylist))

    rellinks = deduplicate(re.findall(link_line, text, flags=re.MULTILINE))
    return rellinks


def _copy_nb(base: str, relpath: str, repo: str, ref: str) -> str:
    """Create a local copy of external file, modifying relative reference"""
    example_dir = Path(__file__).parent / "examples"
    repo_local_dir = example_dir / repo
    repo_local_dir.mkdir(exist_ok=True)
    ref_local_dir = repo_local_dir / ref
    ref_local_dir.mkdir(exist_ok=True)
    page = requests.get(base + relpath)
    if page.status_code != 200:
        raise RuntimeError(f"Unable to locate external notebook at {base + relpath}")
    filename = ref_local_dir / relpath.rsplit("/", 1)[1]
    with open(filename, "wb") as f:
        f.write(page.content)
    return os.path.relpath(filename, start=example_dir)



from . import base
from .base import ConcatLibrary
from .base import TensoredLibrary
from .custom_library import CustomLibrary
from .fourier_library import FourierLibrary
from .generalized_library import GeneralizedLibrary
from .parameterized_library import ParameterizedLibrary
from .pde_library import PDELibrary
from .polynomial_library import IdentityLibrary
from .polynomial_library import PolynomialLibrary
from .sindy_pi_library import SINDyPILibrary
from .weak_pde_library import WeakPDELibrary

__all__ = [
    "ConcatLibrary",
    "TensoredLibrary",
    "GeneralizedLibrary",
    "CustomLibrary",
    "FourierLibrary",
    "IdentityLibrary",
    "PolynomialLibrary",
    "PDELibrary",
    "WeakPDELibrary",
    "SINDyPILibrary",
    "ParameterizedLibrary",
    "base",
]



"""
Unit tests for differentiation methods.
"""
import numpy as np
import pytest
from derivative import dxdt

from pysindy.differentiation import FiniteDifference
from pysindy.differentiation import SINDyDerivative
from pysindy.differentiation import SmoothedFiniteDifference
from pysindy.differentiation import SpectralDerivative
from pysindy.differentiation.base import BaseDifferentiation


@pytest.mark.parametrize(
    "method",
    [
        FiniteDifference(),
        SINDyDerivative(kind="spline", s=1),
        SmoothedFiniteDifference(),
        SpectralDerivative(),
    ],
)
def test_methods_store_smoothed_x_(data_derivative_1d, method):
    x, _ = data_derivative_1d
    assert not hasattr(method, "smoothed_x_")
    method(x)
    assert hasattr(method, "smoothed_x_")
    assert x.shape == method.smoothed_x_.shape


def test_methods_smoothedfd_smooths(data_derivative_1d):
    x, _ = data_derivative_1d
    x_noisy = x + np.random.normal(size=x.shape)
    method = SmoothedFiniteDifference()
    _ = method(x_noisy)
    result = method.smoothed_x_
    assert np.linalg.norm(result - x) < np.linalg.norm(x_noisy - x)


def test_methods_smoothedfd_not_save_smooth(data_derivative_1d):
    x, _ = data_derivative_1d
    method = SmoothedFiniteDifference(save_smooth=False)
    _ = method(x)
    result = method.smoothed_x_
    np.testing.assert_allclose(x, result)


def test_forward_difference_length():
    x = 2 * np.linspace(1, 100, 100)
    forward_difference = FiniteDifference(order=1)
    assert len(forward_difference(x)) == len(x)

    forward_difference_nans = FiniteDifference(order=1, drop_endpoints=True)
    assert len(forward_difference_nans(x)) == len(x)


def test_forward_difference_variable_timestep_length():
    t = np.linspace(1, 10, 100) ** 2
    x = 2 * t
    forward_difference = FiniteDifference(order=1)
    assert len(forward_difference(x, t) == len(x))


def test_centered_difference_length():
    x = 2 * np.linspace(1, 100, 100)
    centered_difference = FiniteDifference(order=2)
    assert len(centered_difference(x)) == len(x)

    centered_difference_nans = FiniteDifference(order=2, drop_endpoints=True)
    assert len(centered_difference_nans(x)) == len(x)


def test_centered_difference_variable_timestep_length():
    t = np.linspace(1, 10, 100) ** 2
    x = 2 * t
    centered_difference = FiniteDifference(order=2)
    assert len(centered_difference(x, t) == len(x))


def test_nan_derivatives(data_lorenz):
    x, t = data_lorenz
    x_dot = FiniteDifference(drop_endpoints=False)(x, t)
    x_dot_nans = FiniteDifference(drop_endpoints=True)(x, t)
    np.testing.assert_allclose(x_dot_nans[1:-1], x_dot[1:-1])
    assert np.isnan(x_dot_nans[:1]).all() and np.isnan(x_dot_nans[-1:]).all()


def test_forward_difference_1d(data_derivative_1d):
    x, x_dot = data_derivative_1d
    forward_difference = FiniteDifference(order=1)
    np.testing.assert_allclose(forward_difference(x), x_dot)


def test_forward_difference_2d(data_derivative_2d):
    x, x_dot = data_derivative_2d
    forward_difference = FiniteDifference(order=1)
    np.testing.assert_allclose(forward_difference(x), x_dot)


def test_centered_difference_1d(data_derivative_1d):
    x, x_dot = data_derivative_1d
    centered_difference = FiniteDifference(order=2)
    np.testing.assert_allclose(centered_difference(x), x_dot)


def test_spectral_derivative_1d(data_derivative_quasiperiodic_1d):
    t, x, x_dot = data_derivative_quasiperiodic_1d
    spectral_derivative = SpectralDerivative()
    np.testing.assert_allclose(spectral_derivative(x, t), x_dot, atol=1e-12)
    centered_difference = FiniteDifference(order=2, periodic=True)._differentiate(x, t)
    np.testing.assert_allclose(
        centered_difference[0], centered_difference[-1], rtol=1e-4
    )


def test_centered_difference_2d(data_derivative_2d):
    x, x_dot = data_derivative_2d
    centered_difference = FiniteDifference(order=2)
    np.testing.assert_allclose(centered_difference(x), x_dot)


def test_spectral_derivative_2d(data_derivative_quasiperiodic_2d):
    t, x, x_dot = data_derivative_quasiperiodic_2d
    spectral_derivative = SpectralDerivative()
    np.testing.assert_allclose(spectral_derivative(x, t), x_dot, atol=1e-12)
    centered_difference = FiniteDifference(order=2, periodic=True)._differentiate(x, t)
    np.testing.assert_allclose(
        centered_difference[0, 0], centered_difference[-1, 0], rtol=1e-4
    )


def test_centered_difference_2d_uniform(data_derivative_2d):
    x, x_dot = data_derivative_2d
    centered_difference = FiniteDifference(order=2, is_uniform=True)
    np.testing.assert_allclose(centered_difference(x), x_dot)


def test_centered_difference_2d_uniform_time(data_derivative_2d):
    x, x_dot = data_derivative_2d
    t = np.linspace(0, x.shape[0] - 1, x.shape[0])
    centered_difference = FiniteDifference(order=2)
    np.testing.assert_allclose(centered_difference(x, t), x_dot)


def test_centered_difference_2d_nonuniform_time(data_derivative_2d):
    x, x_dot = data_derivative_2d
    t = np.linspace(0, x.shape[0] - 1, x.shape[0])
    centered_difference = FiniteDifference(order=2)
    np.testing.assert_allclose(centered_difference(x, t), x_dot, atol=1e-8)


def test_centered_difference_xy_yx(data_2dspatial):
    x_grid, y_grid, u = data_2dspatial
    u_xy = np.zeros(u.shape)
    u_yx = np.zeros(u.shape)
    u_y = FiniteDifference(order=2, d=1, axis=1)._differentiate(u, y_grid)
    u_x = FiniteDifference(order=2, d=1, axis=0)._differentiate(u, x_grid)
    u_xy = FiniteDifference(order=2, d=1, axis=1)._differentiate(u_x, y_grid)
    u_yx = FiniteDifference(order=2, d=1, axis=0)._differentiate(u_y, x_grid)
    np.testing.assert_allclose(u_xy, u_yx)
    u_y = FiniteDifference(order=1, d=1, axis=1)._differentiate(u, y_grid)
    u_x = FiniteDifference(order=1, d=1, axis=0)._differentiate(u, x_grid)
    u_xy = FiniteDifference(order=1, d=1, axis=1)._differentiate(u_x, y_grid)
    u_yx = FiniteDifference(order=1, d=1, axis=0)._differentiate(u_y, x_grid)
    np.testing.assert_allclose(u_xy, u_yx)


def test_centered_difference_hot(data_derivative_2d):
    x, _ = data_derivative_2d
    t = np.linspace(0, x.shape[0], x.shape[0])
    dt = t[1] - t[0]
    atol = 1e-8
    for d in range(1, 2):
        forward_difference = FiniteDifference(order=1, d=d)._differentiate
        np.testing.assert_allclose(
            forward_difference(x, t=dt),
            forward_difference(x, t=t),
            atol=atol,
        )
    for d in range(1, 6):
        centered_difference = FiniteDifference(order=2, d=d)._differentiate
        np.testing.assert_allclose(
            centered_difference(x, t=dt),
            centered_difference(x, t=t),
            atol=atol,
        )
    for d in range(1, 6):
        spectral_deriv = SpectralDerivative(d=d)._differentiate
        np.testing.assert_allclose(
            spectral_deriv(x, t=dt),
            spectral_deriv(x, t=t),
            atol=atol,
        )


# Alternative implementation of the four tests above using parametrization
@pytest.mark.parametrize(
    "data, order",
    [
        (pytest.lazy_fixture("data_derivative_1d"), 2),
        (pytest.lazy_fixture("data_derivative_2d"), 2),
        (pytest.lazy_fixture("data_derivative_1d"), 4),
        (pytest.lazy_fixture("data_derivative_2d"), 4),
        (pytest.lazy_fixture("data_derivative_1d"), 8),
        (pytest.lazy_fixture("data_derivative_2d"), 8),
    ],
)
def test_finite_difference(data, order):
    x, x_dot = data
    method = FiniteDifference(order=order)
    np.testing.assert_allclose(method(x), x_dot)
    method = SmoothedFiniteDifference()
    np.testing.assert_allclose(method(x), x_dot)


def test_order_error():
    with pytest.raises(ValueError):
        FiniteDifference(order=-1)
    with pytest.raises(ValueError):
        FiniteDifference(d=-1)
    with pytest.raises(ValueError):
        FiniteDifference(d=2, order=1)
    with pytest.raises(ValueError):
        FiniteDifference(d=1, order=0.5)
    with pytest.raises(ValueError):
        FiniteDifference(d=1, order=0)
    with pytest.raises(ValueError):
        FiniteDifference(d=0, order=1)


def test_base_class(data_derivative_1d):
    x, x_dot = data_derivative_1d
    with pytest.raises(NotImplementedError):
        BaseDifferentiation()._differentiate(x)


# Test smoothed finite difference method
@pytest.mark.parametrize(
    "data",
    [
        pytest.lazy_fixture("data_derivative_1d"),
        pytest.lazy_fixture("data_derivative_2d"),
    ],
)
def test_smoothed_finite_difference(data):
    x, x_dot = data
    smoothed_centered_difference = SmoothedFiniteDifference()
    np.testing.assert_allclose(smoothed_centered_difference(x), x_dot)


@pytest.mark.parametrize(
    "data, derivative_kws",
    [
        (pytest.lazy_fixture("data_derivative_1d"), dict(kind="spectral")),
        (pytest.lazy_fixture("data_derivative_2d"), dict(kind="spectral")),
        (pytest.lazy_fixture("data_derivative_1d"), dict(kind="spline", s=1e-2)),
        (pytest.lazy_fixture("data_derivative_2d"), dict(kind="spline", s=1e-2)),
        (
            pytest.lazy_fixture("data_derivative_1d"),
            dict(kind="trend_filtered", order=0, alpha=1e-2),
        ),
        (
            pytest.lazy_fixture("data_derivative_2d"),
            dict(kind="trend_filtered", order=0, alpha=1e-2),
        ),
        (
            pytest.lazy_fixture("data_derivative_1d"),
            dict(kind="finite_difference", k=1),
        ),
        (
            pytest.lazy_fixture("data_derivative_2d"),
            dict(kind="finite_difference", k=1),
        ),
        (
            pytest.lazy_fixture("data_derivative_1d"),
            dict(kind="savitzky_golay", order=3, left=1, right=1),
        ),
        (
            pytest.lazy_fixture("data_derivative_2d"),
            dict(kind="savitzky_golay", order=3, left=1, right=1),
        ),
    ],
)
def test_wrapper_equivalence_with_dxdt(data, derivative_kws):
    x, _ = data
    t = np.arange(x.shape[0])

    if np.ndim(x) == 1:
        np.testing.assert_allclose(
            dxdt(x, t, axis=0, **derivative_kws),
            SINDyDerivative(**derivative_kws)(x, t),
        )
    else:
        np.testing.assert_allclose(
            dxdt(x, t, axis=0, **derivative_kws),
            SINDyDerivative(**derivative_kws)(x, t),
        )


def test_sindy_derivative_kwarg_update():
    method = SINDyDerivative(kind="spectral", foo=2)
    method.set_params(kwargs={"kind": "spline", "foo": 1})
    assert method.kwargs["kind"] == "spline"
    assert method.kwargs["foo"] == 1


@pytest.mark.parametrize(
    "data, derivative_kws",
    [
        (pytest.lazy_fixture("data_derivative_1d"), dict(kind="spectral")),
        (pytest.lazy_fixture("data_derivative_2d"), dict(kind="spectral")),
        (pytest.lazy_fixture("data_derivative_1d"), dict(kind="spline", s=1e-2)),
        (pytest.lazy_fixture("data_derivative_2d"), dict(kind="spline", s=1e-2)),
        (
            pytest.lazy_fixture("data_derivative_1d"),
            dict(kind="trend_filtered", order=0, alpha=1e-2),
        ),
        (
            pytest.lazy_fixture("data_derivative_2d"),
            dict(kind="trend_filtered", order=0, alpha=1e-2),
        ),
        (
            pytest.lazy_fixture("data_derivative_1d"),
            dict(kind="finite_difference", k=1),
        ),
        (
            pytest.lazy_fixture("data_derivative_2d"),
            dict(kind="finite_difference", k=1),
        ),
        (
            pytest.lazy_fixture("data_derivative_1d"),
            dict(kind="savitzky_golay", order=3, left=1, right=1),
        ),
        (
            pytest.lazy_fixture("data_derivative_2d"),
            dict(kind="savitzky_golay", order=3, left=1, right=1),
        ),
    ],
)
def test_derivative_output_shape(data, derivative_kws):
    x, x_dot = data
    t = np.arange(x.shape[0])

    method = SINDyDerivative(**derivative_kws)

    assert x_dot.shape == method(x).shape
    assert x_dot.shape == method(x, t).shape


def test_bad_t_values(data_derivative_1d):
    x, x_dot = data_derivative_1d

    method = SINDyDerivative(kind="finite_difference", k=1)

    with pytest.raises(ValueError):
        method(x, t=-1)

    with pytest.raises(ValueError):
        method._differentiate(x, t=-1)


def test_centered_difference_hot_axis(data_2d_resolved_pde):
    _, u_flat, u_dot_flat = data_2d_resolved_pde
    x = np.reshape(u_flat, (8, 8, 1000, 2))
    t = np.linspace(0, 10, 1000)
    dt = t[1] - t[0]
    atol = 1e-8
    for d in range(1, 2):
        forward_difference = FiniteDifference(order=1, d=d, axis=-2)._differentiate
        np.testing.assert_allclose(
            forward_difference(x, t=dt),
            forward_difference(x, t=t),
            atol=atol,
        )
    for d in range(1, 6):
        centered_difference = FiniteDifference(order=2, d=d, axis=-2)._differentiate
        np.testing.assert_allclose(
            centered_difference(x, t=dt),
            centered_difference(x, t=t),
            atol=atol,
        )
    for d in range(1, 6):
        spectral_deriv = SpectralDerivative(d=d, axis=-2)._differentiate
        np.testing.assert_allclose(
            spectral_deriv(x, t=dt),
            spectral_deriv(x, t=t),
            atol=atol,
        )


def test_centered_difference_noaxis_vs_axis(data_2d_resolved_pde):
    _, u_flat, u_dot_flat = data_2d_resolved_pde
    n = 8
    x = np.reshape(u_flat, (n, n, 1000, 2))
    t = np.linspace(0, 10, 1000)
    dt = t[1] - t[0]
    atol = 1e-10
    for d in range(1, 6):
        centered_difference = FiniteDifference(order=2, d=d, axis=-2)._differentiate
        slow_differences = np.zeros(x.shape)
        slow_differences_t = np.zeros(x.shape)
        for i in range(n):
            for j in range(n):
                slow_differences[i, j, :, :] = FiniteDifference(
                    order=2, d=d
                )._differentiate(x[i, j, :, :], t=dt)
                slow_differences_t[i, j, :, :] = FiniteDifference(
                    order=2, d=d
                )._differentiate(x[i, j, :, :], t=t)
        np.testing.assert_allclose(
            centered_difference(x, t=dt),
            slow_differences,
            atol=atol,
        )
        np.testing.assert_allclose(
            centered_difference(x, t=t),
            slow_differences_t,
            atol=atol,
        )


@pytest.mark.parametrize(
    "derivative, kwargs",
    [
        (SINDyDerivative, {"kind": "finite_difference", "k": 1, "axis": -2}),
        (FiniteDifference, {"axis": -2}),
        (
            SmoothedFiniteDifference,
            {"axis": -2, "smoother_kws": {"window_length": 2, "polyorder": 1}},
        ),
        (SpectralDerivative, {"axis": -2}),
    ],
)
def test_nd_differentiation(derivative, kwargs):
    t = np.arange(3)
    x = np.random.random(size=(2, 3, 2))
    x[1, :, 1] = 1
    xdot = derivative(**kwargs)._differentiate(x, t)
    expected = np.zeros(3)
    np.testing.assert_array_almost_equal(xdot[1, :, 1], expected)



import pytest
from numpy.testing import assert_allclose
from sklearn.exceptions import NotFittedError

from pysindy import FourierLibrary
from pysindy import SINDy
from pysindy import STLSQ
from pysindy.deeptime import SINDyEstimator
from pysindy.deeptime import SINDyModel


def test_estimator_has_model(data_lorenz):
    x, t = data_lorenz

    estimator = SINDyEstimator()
    assert not estimator.has_model

    estimator.fit(x, t=t)
    assert estimator.has_model


def test_estimator_fetch_model(data_lorenz):
    x, t = data_lorenz

    estimator = SINDyEstimator()
    assert estimator.fetch_model() is None

    estimator.fit(x, t=t)
    assert isinstance(estimator.fetch_model(), SINDyModel)


def test_model_sindy_equivalence(data_lorenz_c_1d):
    x, t, u, _ = data_lorenz_c_1d

    model = SINDyEstimator().fit(x, t=t, u=u).fetch_model()
    sindy_model = SINDy().fit(x, t=t, u=u)

    assert_allclose(model.coefficients(), sindy_model.coefficients())
    print(sindy_model.n_features_in_)
    assert model.n_features_in_ == sindy_model.n_features_in_
    assert model.n_output_features_ == sindy_model.n_output_features_
    assert model.n_control_features_ == sindy_model.n_control_features_


def test_model_has_sindy_methods(data_lorenz):
    x, t = data_lorenz
    model = SINDyEstimator().fit(x, t=t).fetch_model()

    assert hasattr(model, "predict")
    assert hasattr(model, "simulate")
    assert hasattr(model, "score")
    assert hasattr(model, "print")
    assert hasattr(model, "equations")


def test_model_unfitted_library(data_derivative_2d):
    x, x_dot = data_derivative_2d
    optimizer = STLSQ().fit(x, x_dot)
    library = FourierLibrary()

    with pytest.raises(NotFittedError):
        SINDyModel(optimizer, library)


def test_model_unfitted_optimizer(data_lorenz):
    x, t = data_lorenz
    optimizer = STLSQ()
    library = FourierLibrary().fit(x)

    with pytest.raises(NotFittedError):
        SINDyModel(optimizer, library)


def test_model_copy(data_lorenz):
    x, t = data_lorenz
    model = SINDyEstimator().fit(x, t=t).fetch_model()
    model_copy = model.copy()

    assert model is not model_copy



"""
Unit tests for feature libraries.
"""
import numpy as np
import pytest
from scipy.sparse import csr_matrix
from sklearn.exceptions import NotFittedError
from sklearn.utils.validation import check_is_fitted

from pysindy import SINDy
from pysindy.feature_library import ConcatLibrary
from pysindy.feature_library import CustomLibrary
from pysindy.feature_library import FourierLibrary
from pysindy.feature_library import GeneralizedLibrary
from pysindy.feature_library import IdentityLibrary
from pysindy.feature_library import ParameterizedLibrary
from pysindy.feature_library import PDELibrary
from pysindy.feature_library import PolynomialLibrary
from pysindy.feature_library import SINDyPILibrary
from pysindy.feature_library import TensoredLibrary
from pysindy.feature_library import WeakPDELibrary
from pysindy.feature_library.base import BaseFeatureLibrary
from pysindy.feature_library.polynomial_library import n_poly_features
from pysindy.optimizers import SINDyPI
from pysindy.optimizers import STLSQ


def test_form_custom_library():
    library_functions = [lambda x: x, lambda x: x**2, lambda x: 0 * x]
    function_names = [
        lambda s: str(s),
        lambda s: "{}^2".format(s),
        lambda s: "0",
    ]

    # Test with user-supplied function names
    CustomLibrary(library_functions=library_functions, function_names=function_names)

    # Test without user-supplied function names
    CustomLibrary(library_functions=library_functions, function_names=None)


def test_form_pde_library():
    function_library = PolynomialLibrary(degree=3, include_bias=False)

    # Test with user-supplied function names
    PDELibrary(function_library=function_library)


def test_form_sindy_pi_library():
    library_functions = [lambda x: x, lambda x: x**2, lambda x: 0 * x]
    function_names = [
        lambda s: str(s),
        lambda s: "{}^2".format(s),
        lambda s: "0",
    ]
    # Test with user-supplied function names
    SINDyPILibrary(library_functions=library_functions, function_names=function_names)

    # Test without user-supplied function names
    SINDyPILibrary(library_functions=library_functions, function_names=None)


def test_bad_parameters():
    with pytest.raises(ValueError):
        PolynomialLibrary(degree=-1).fit(np.array([]))
    with pytest.raises(ValueError):
        PolynomialLibrary(degree=1.5).fit(np.array([]))
    with pytest.raises(ValueError):
        PolynomialLibrary(include_interaction=False, interaction_only=True).fit(
            np.array([])
        )
    with pytest.raises(ValueError):
        FourierLibrary(n_frequencies=-1)
    with pytest.raises(ValueError):
        FourierLibrary(n_frequencies=-1)
    with pytest.raises(ValueError):
        FourierLibrary(n_frequencies=2.2)
    with pytest.raises(ValueError):
        FourierLibrary(include_sin=False, include_cos=False)
    with pytest.raises(ValueError):
        library_functions = [lambda x: x, lambda x: x**2, lambda x: 0 * x]
        function_names = [lambda s: str(s), lambda s: "{}^2".format(s)]
        CustomLibrary(
            library_functions=library_functions, function_names=function_names
        )


@pytest.mark.parametrize(
    "params",
    [
        dict(derivative_order=1),
        dict(derivative_order=3),
        dict(spatial_grid=range(10)),
        dict(spatial_grid=range(10), derivative_order=-1),
        dict(spatial_grid=np.zeros((10, 10))),
        dict(spatial_grid=np.zeros((10, 10, 10, 10, 10))),
    ],
)
def test_pde_library_bad_parameters(params):
    with pytest.raises(ValueError):
        PDELibrary(**params)


@pytest.mark.parametrize(
    "params",
    [
        dict(spatiotemporal_grid=range(10), p=-1),
        dict(spatiotemporal_grid=range(10), H_xt=-1),
        dict(spatiotemporal_grid=range(10), H_xt=11),
        dict(spatiotemporal_grid=range(10), K=-1),
        dict(),
        dict(
            spatiotemporal_grid=np.asarray(np.meshgrid(range(10), range(10))).T, H_xt=-1
        ),
        dict(
            spatiotemporal_grid=np.transpose(
                np.asarray(np.meshgrid(range(10), range(10), range(10), indexing="ij")),
                axes=[1, 2, 3, 0],
            ),
            H_xt=-1,
        ),
        dict(
            spatiotemporal_grid=np.transpose(
                np.asarray(np.meshgrid(range(10), range(10), range(10), indexing="ij")),
                axes=[1, 2, 3, 0],
            ),
            H_xt=11,
        ),
    ],
)
def test_weak_pde_library_bad_parameters(params):
    with pytest.raises(ValueError):
        WeakPDELibrary(**params)


@pytest.mark.parametrize(
    "params",
    [
        dict(libraries=[]),
        dict(
            libraries=[PolynomialLibrary(), PolynomialLibrary()], tensor_array=[[0, 0]]
        ),
        dict(
            libraries=[PolynomialLibrary(), PolynomialLibrary()], tensor_array=[[0, 1]]
        ),
        dict(
            libraries=[PolynomialLibrary(), PolynomialLibrary()], tensor_array=[[1, -1]]
        ),
        dict(
            libraries=[PolynomialLibrary(), PolynomialLibrary()], tensor_array=[[2, 1]]
        ),
        dict(libraries=[PolynomialLibrary(), PolynomialLibrary()], tensor_array=[1, 1]),
        dict(
            libraries=[PolynomialLibrary(), PolynomialLibrary()],
            tensor_array=[[1, 1, 1]],
        ),
        dict(
            libraries=[PolynomialLibrary(), PolynomialLibrary()],
            inputs_per_library=[[0, 1], [0, 100]],
        ),
        dict(
            libraries=[PolynomialLibrary(), PolynomialLibrary()],
            inputs_per_library=[[0, 1]],
        ),
        dict(
            libraries=[PolynomialLibrary(), PolynomialLibrary()],
            inputs_per_library=[[0, 1, 2], [0, 1, 2], [0, 1, 2]],
        ),
        dict(
            libraries=[PolynomialLibrary(), PolynomialLibrary()],
            inputs_per_library=[[0, 1, 2], [0, 1, -1]],
        ),
    ],
)
def test_generalized_library_bad_parameters(data_lorenz, params):
    with pytest.raises(ValueError):
        lib = GeneralizedLibrary(**params)
        x, t = data_lorenz
        lib.fit(x)


@pytest.mark.parametrize(
    "params",
    [
        dict(num_parameters=0, num_features=1),
        dict(num_parameters=1, num_features=0),
        dict(feature_library=None),
        dict(parameter_library=None),
    ],
)
def test_parameterized_library_bad_parameters(data_lorenz, params):
    with pytest.raises(ValueError):
        lib = ParameterizedLibrary(**params)
        x, t = data_lorenz
        lib.fit(x)


@pytest.mark.parametrize(
    "params",
    [
        dict(
            library_functions=[lambda x: x, lambda x: x**2, lambda x: 0 * x],
            function_names=[lambda s: str(s), lambda s: "{}^2".format(s)],
        ),
        dict(
            x_dot_library_functions=[lambda x: x, lambda x: x**2, lambda x: 0 * x],
            function_names=[lambda s: str(s), lambda s: "{}^2".format(s)],
        ),
        dict(x_dot_library_functions=[lambda x: x, lambda x: x**2, lambda x: 0 * x]),
        dict(),
        dict(
            library_functions=[lambda x: x, lambda x: x**2],
            x_dot_library_functions=[lambda x: x, lambda x: x**2],
            function_names=[lambda s: s, lambda s: s + s],
        ),
    ],
)
def test_sindypi_library_bad_params(params):
    with pytest.raises(ValueError):
        SINDyPILibrary(**params)


@pytest.mark.parametrize(
    "library",
    [
        PolynomialLibrary(),
        PolynomialLibrary(include_bias=False),
        FourierLibrary(),
        IdentityLibrary() + PolynomialLibrary(),
        pytest.lazy_fixture("custom_library"),
        pytest.lazy_fixture("custom_library_bias"),
        pytest.lazy_fixture("generalized_library"),
        pytest.lazy_fixture("ode_library"),
        pytest.lazy_fixture("sindypi_library"),
    ],
    ids=type,
)
def test_fit_transform(data_lorenz, library):
    x, t = data_lorenz
    library.fit_transform(x)
    check_is_fitted(library)


@pytest.mark.parametrize(
    "library",
    [
        PolynomialLibrary(),
        PolynomialLibrary(include_bias=False),
        FourierLibrary(),
        IdentityLibrary() + PolynomialLibrary(),
        pytest.lazy_fixture("custom_library"),
        pytest.lazy_fixture("custom_library_bias"),
        pytest.lazy_fixture("generalized_library"),
        pytest.lazy_fixture("ode_library"),
        pytest.lazy_fixture("pde_library"),
        pytest.lazy_fixture("sindypi_library"),
    ],
)
def test_change_in_data_shape(data_lorenz, library):
    x, t = data_lorenz
    library.fit(x)
    with pytest.raises(ValueError):
        library.transform(x[:, 1:])


@pytest.mark.parametrize(
    "library, shape",
    [
        (PolynomialLibrary(include_bias=False), 9),
        (PolynomialLibrary(), 10),
        (IdentityLibrary() + PolynomialLibrary(), 13),
        (FourierLibrary(), 6),
        (pytest.lazy_fixture("custom_library_bias"), 13),
        (pytest.lazy_fixture("custom_library"), 12),
        (pytest.lazy_fixture("generalized_library"), 76),
        (pytest.lazy_fixture("ode_library"), 9),
        (pytest.lazy_fixture("sindypi_library"), 39),
    ],
)
def test_output_shape(data_lorenz, library, shape):
    x, t = data_lorenz
    y = library.fit_transform(x)
    expected_shape = (x.shape[0], shape)
    assert y.shape == expected_shape
    assert library.size > 0


@pytest.mark.parametrize(
    "library",
    [
        PolynomialLibrary(),
        PolynomialLibrary(include_bias=False),
        FourierLibrary(),
        PolynomialLibrary() + FourierLibrary(),
        pytest.lazy_fixture("custom_library"),
        pytest.lazy_fixture("custom_library_bias"),
        pytest.lazy_fixture("generalized_library"),
        pytest.lazy_fixture("ode_library"),
        pytest.lazy_fixture("sindypi_library"),
    ],
)
def test_get_feature_names(data_lorenz, library):
    with pytest.raises(NotFittedError):
        library.get_feature_names()

    x, t = data_lorenz
    library.fit_transform(x)
    feature_names = library.get_feature_names()
    assert isinstance(feature_names, list)
    assert isinstance(feature_names[0], str)

    input_features = ["a"] * x.shape[1]
    library.get_feature_names(input_features=input_features)
    assert isinstance(feature_names, list)
    assert isinstance(feature_names[0], str)


# Catch-all for various combinations of options and
# inputs for polynomial features
@pytest.mark.parametrize(
    "kwargs, sparse_format",
    [
        ({"degree": 4}, csr_matrix),
        ({"include_bias": True}, csr_matrix),
        ({"include_bias": False}, csr_matrix),
        ({"include_bias": False, "interaction_only": True}, csr_matrix),
        (
            {
                "include_bias": False,
                "interaction_only": False,
                "include_interaction": False,
            },
            csr_matrix,
        ),
        (
            {
                "include_bias": False,
                "interaction_only": False,
                "include_interaction": True,
            },
            csr_matrix,
        ),
        ({"include_interaction": False}, lambda x: x),
        ({"include_interaction": False, "include_bias": True}, lambda x: x),
    ],
)
def test_polynomial_options(data_lorenz, kwargs, sparse_format):
    x, t = data_lorenz
    library = PolynomialLibrary(**kwargs)
    out = library.fit_transform(sparse_format(x))
    check_is_fitted(library)
    expected = len(library.powers_)
    result = out.shape[1]
    assert result == expected


# Catch-all for various combinations of options and
# inputs for Fourier features
def test_fourier_options(data_lorenz):
    x, t = data_lorenz

    library = FourierLibrary(include_cos=False)
    library.fit_transform(x)
    check_is_fitted(library)


def test_not_implemented(data_lorenz):
    x, t = data_lorenz
    library = BaseFeatureLibrary()

    with pytest.raises(NotImplementedError):
        library.fit(x)

    with pytest.raises(NotImplementedError):
        library.transform(x)

    with pytest.raises(NotImplementedError):
        library.get_feature_names(x)


def test_concat(data_lorenz):
    x, t = data_lorenz
    ident_lib = IdentityLibrary()
    poly_lib = PolynomialLibrary()
    concat_lib = ident_lib + poly_lib
    assert isinstance(concat_lib, ConcatLibrary)
    concat_lib.fit(x)
    check_is_fitted(concat_lib)
    concat_lib.fit_transform(x)
    check_is_fitted(concat_lib)


def test_tensored(data_lorenz):
    x, t = data_lorenz
    ident_lib = IdentityLibrary()
    poly_lib = PolynomialLibrary()
    tensored_lib = ident_lib * poly_lib
    assert isinstance(tensored_lib, TensoredLibrary)
    tensored_lib.fit(x)
    check_is_fitted(tensored_lib)
    tensored_lib.fit_transform(x)
    check_is_fitted(tensored_lib)


@pytest.mark.parametrize(
    "library",
    [
        PolynomialLibrary(),
        FourierLibrary(),
        PolynomialLibrary() + FourierLibrary(),
        pytest.lazy_fixture("custom_library"),
        pytest.lazy_fixture("generalized_library"),
        pytest.lazy_fixture("ode_library"),
        pytest.lazy_fixture("pde_library"),
        pytest.lazy_fixture("sindypi_library"),
    ],
)
def test_not_fitted(data_lorenz, library):
    x, t = data_lorenz

    with pytest.raises(NotFittedError):
        library.transform(x)


def test_tensored_library():
    f_list = [lambda x: x]
    lib_names_a = [lambda x: f"a({x})"]
    lib_names_b = [lambda x: f"b({x})"]
    lib_names_c = [lambda x: f"c({x})"]

    lib_a = CustomLibrary(f_list, lib_names_a)
    lib_b = CustomLibrary(f_list, lib_names_b)
    lib_c = CustomLibrary(f_list, lib_names_c)

    libraries = [lib_a, lib_b, lib_c]

    inputs_per_library = [[1], [1], [0]]
    x = np.array([[1, 2]])

    tensored_library = np.prod(np.asarray(libraries))
    tensored_library._set_inputs_per_library(inputs_per_library)

    values = tensored_library.fit_transform(x)
    names = tensored_library.get_feature_names(input_features=["x", "y"])

    assert names == ["a(y) b(y) c(x)"]
    assert values.item() == 4


def test_generalized_library(data_lorenz):
    x, t = data_lorenz
    poly_library = PolynomialLibrary(include_bias=False)
    fourier_library = FourierLibrary()
    library_functions = [
        lambda x: np.exp(x),
        lambda x: 1.0 / x,
        lambda x, y: np.sin(x + y),
    ]
    custom_library = CustomLibrary(
        library_functions=library_functions,
    )

    tensor_array = [[0, 1, 1], [1, 0, 1]]

    inputs_per_library = [[1, 2], [0, 2], [0]]

    # First try without tensor libraries and subset of the input variables
    sindy_library = GeneralizedLibrary(
        [poly_library, fourier_library, custom_library],
    )
    sindy_opt = STLSQ(threshold=0.25)
    model = SINDy(
        optimizer=sindy_opt,
        feature_library=sindy_library,
    )
    model.fit(x, t=t)
    model.print()
    model.get_feature_names()
    assert len(model.get_feature_names()) == 24

    # Repeat with feature names
    feature_names = ["x", "y", "z"]
    model = SINDy(optimizer=sindy_opt, feature_library=sindy_library)
    model.fit(x, t=t, feature_names=feature_names)
    model.print()
    model.get_feature_names()
    assert len(model.get_feature_names()) == 24

    # Next try with tensor libraries but still all the input variables
    sindy_library = GeneralizedLibrary(
        [poly_library, fourier_library, custom_library], tensor_array=tensor_array
    )
    model = SINDy(
        optimizer=sindy_opt,
        feature_library=sindy_library,
    )
    model.fit(x, t=t)
    model.print()
    # 24 + (9 * 6) = 54 + (9 * 9) = 81
    assert len(model.get_feature_names()) == 159

    # Repeat with feature_names
    sindy_library = GeneralizedLibrary(
        [poly_library, fourier_library, custom_library], tensor_array=tensor_array
    )
    model = SINDy(optimizer=sindy_opt, feature_library=sindy_library)
    model.fit(x, t=t, feature_names=feature_names)
    model.print()
    # 24 + (9 * 6) = 54 + (9 * 9) = 81
    assert len(model.get_feature_names()) == 159

    sindy_library = GeneralizedLibrary(
        [poly_library, fourier_library, custom_library],
        tensor_array=tensor_array,
        inputs_per_library=inputs_per_library,
    )
    model = SINDy(
        optimizer=sindy_opt,
        feature_library=sindy_library,
    )
    model.fit(x, t=t)
    assert len(model.get_feature_names()) == 29

    # Repeat with feature names
    sindy_library = GeneralizedLibrary(
        [poly_library, fourier_library, custom_library],
        tensor_array=tensor_array,
        inputs_per_library=inputs_per_library,
    )
    model = SINDy(optimizer=sindy_opt, feature_library=sindy_library)
    model.fit(x, t=t, feature_names=feature_names)
    assert len(model.get_feature_names()) == 29


def test_generalized_library_pde(data_1d_random_pde):
    t, x, u, u_dot = data_1d_random_pde
    poly_library = PolynomialLibrary(include_bias=False)
    fourier_library = FourierLibrary()
    pde_library = PDELibrary(
        function_library=PolynomialLibrary(degree=2, include_bias=False),
        derivative_order=2,
        spatial_grid=x,
        include_bias=True,
    )

    # First try without tensor libraries and subset of the input variables
    sindy_library = GeneralizedLibrary(
        [poly_library, fourier_library, pde_library],
    )
    sindy_opt = STLSQ(threshold=0.25)
    model = SINDy(
        optimizer=sindy_opt,
        feature_library=sindy_library,
    )
    model.fit(u, t=t)
    model.print()
    model.get_feature_names()
    assert len(model.get_feature_names()) == 13


def test_generalized_library_weak_pde(data_1d_random_pde):
    t, x, u, u_dot = data_1d_random_pde
    X, T = np.meshgrid(x, t)
    XT = np.array([X, T]).T
    weak_library1 = WeakPDELibrary(
        function_library=PolynomialLibrary(degree=2, include_bias=False),
        derivative_order=2,
        spatiotemporal_grid=XT,
        include_bias=True,
    )
    weak_library2 = WeakPDELibrary(
        function_library=CustomLibrary(library_functions=[lambda x: x * x * x]),
        derivative_order=0,
        spatiotemporal_grid=XT,
    )

    # First try without tensor libraries and subset of the input variables
    sindy_library = GeneralizedLibrary(
        [weak_library1, weak_library2],
    )
    sindy_opt = STLSQ(threshold=0.25)
    model = SINDy(
        optimizer=sindy_opt,
        feature_library=sindy_library,
    )
    model.fit(u, t=t)
    model.print()
    model.get_feature_names()
    assert len(model.get_feature_names()) == 10


def test_parameterized_library(diffuse_multiple_trajectories):
    t, spatial_grid, xs = diffuse_multiple_trajectories
    us = []
    for i in range(len(xs)):
        u = np.zeros(xs[0].shape)
        us = us + [u]

    feature_lib = PDELibrary(
        function_library=PolynomialLibrary(degree=1, include_bias=False),
        derivative_order=2,
        spatial_grid=spatial_grid,
    )

    parameter_lib = PDELibrary(
        function_library=PolynomialLibrary(degree=1, include_bias=False),
        derivative_order=0,
        include_bias=True,
    )

    pde_lib = ParameterizedLibrary(
        feature_library=feature_lib,
        parameter_library=parameter_lib,
        num_features=1,
        num_parameters=1,
    )

    X, T = np.meshgrid(spatial_grid, t, indexing="ij")
    XT = np.transpose([X, T], [1, 2, 0])

    np.random.seed(100)
    weak_feature_lib = WeakPDELibrary(
        function_library=PolynomialLibrary(degree=1, include_bias=False),
        derivative_order=2,
        spatiotemporal_grid=XT,
        K=100,
    )
    np.random.seed(100)
    weak_parameter_lib = WeakPDELibrary(
        function_library=PolynomialLibrary(degree=1, include_bias=False),
        derivative_order=0,
        spatiotemporal_grid=XT,
        K=100,
        include_bias=True,
    )

    weak_lib = ParameterizedLibrary(
        feature_library=weak_feature_lib,
        parameter_library=weak_parameter_lib,
        num_features=1,
        num_parameters=1,
    )

    optimizer = STLSQ(threshold=0.5, alpha=1e-8, normalize_columns=False)
    model = SINDy(feature_library=pde_lib, optimizer=optimizer)
    model.fit(xs, u=us, t=t, feature_names=["u", "c"])
    assert abs(model.coefficients()[0, 4] - 1) < 1e-1
    assert np.all(model.coefficients()[0, :4] == 0)
    assert np.all(model.coefficients()[0, 5:] == 0)

    optimizer = STLSQ(threshold=0.25, alpha=1e-8, normalize_columns=False)
    model = SINDy(feature_library=weak_lib, optimizer=optimizer)
    model.fit(xs, u=us, t=t, feature_names=["u", "c"])
    assert abs(model.coefficients()[0, 4] - 1) < 1e-1
    assert np.all(model.coefficients()[0, :4] == 0)
    assert np.all(model.coefficients()[0, 5:] == 0)


# Helper function for testing PDE libraries
def pde_library_helper(library, u, t):
    base_opt = STLSQ(normalize_columns=True, alpha=1e-10, threshold=0)
    model = SINDy(optimizer=base_opt, feature_library=library)
    model.fit(u, t)
    assert np.any(base_opt.coef_ != 0.0)


def test_1D_pdes(data_1d_random_pde):
    t, spatial_grid, u, _ = data_1d_random_pde
    pde_lib = PDELibrary(
        function_library=PolynomialLibrary(degree=2, include_bias=False),
        derivative_order=4,
        spatial_grid=spatial_grid,
        include_bias=True,
    )
    pde_library_helper(pde_lib, u, t)


def test_1D_weak_pdes():
    n = 10
    t = np.linspace(0, 10, n)
    x = np.linspace(0, 10, n)
    u = np.random.randn(n, n, 1)
    X, T = np.meshgrid(x, t, indexing="ij")
    spatiotemporal_grid = np.asarray([X, T])
    spatiotemporal_grid = np.transpose(spatiotemporal_grid, axes=[1, 2, 0])
    pde_lib = WeakPDELibrary(
        function_library=PolynomialLibrary(degree=2, include_bias=False),
        derivative_order=4,
        spatiotemporal_grid=spatiotemporal_grid,
        H_xt=2,
        include_bias=True,
    )
    pde_library_helper(pde_lib, u, t)


def test_sindypi_library(data_lorenz):
    x, t = data_lorenz
    x_library_functions = [
        lambda x: x,
        lambda x, y: x * y,
        lambda x: x**2,
    ]
    x_dot_library_functions = [lambda x: x]

    library_function_names = [
        lambda x: x,
        lambda x, y: x + y,
        lambda x: x + x,
        lambda x: x,
    ]
    sindy_library = SINDyPILibrary(
        library_functions=x_library_functions,
        x_dot_library_functions=x_dot_library_functions,
        t=t,
        function_names=library_function_names,
        include_bias=True,
    )
    sindy_opt = SINDyPI(reg_weight_lam=0.1, regularizer="l1")
    model = SINDy(
        optimizer=sindy_opt,
        feature_library=sindy_library,
    )
    model.fit(x, t=t)
    assert np.shape(sindy_opt.coef_) == (40, 40)

    sindy_opt = SINDyPI(reg_weight_lam=1, regularizer="l1", model_subset=[3])
    model = SINDy(
        optimizer=sindy_opt,
        feature_library=sindy_library,
    )
    model.fit(x, t=t)
    assert np.sum(sindy_opt.coef_ == 0.0) == 40.0 * 39.0 and np.any(
        sindy_opt.coef_[3, :] != 0.0
    )


@pytest.mark.parametrize(
    ("include_interaction", "interaction_only", "bias", "expected"),
    [
        (True, True, True, ((), (0,), (0, 1), (1,))),
        (False, False, False, ((0,), (0, 0), (1,), (1, 1))),
    ],
)
def test_polynomial_combinations(include_interaction, interaction_only, bias, expected):
    combos = PolynomialLibrary._combinations(
        n_features=2,
        degree=2,
        include_interaction=include_interaction,
        interaction_only=interaction_only,
        include_bias=bias,
    )
    result = tuple(sorted(list(combos)))
    assert result == expected
    assert len(expected) == n_poly_features(
        2, 2, bias, include_interaction, interaction_only
    )



import numpy as np
from sklearn.multioutput import MultiOutputRegressor

from .base import BaseOptimizer

COEF_THRESHOLD = 1e-14


class WrappedOptimizer(BaseOptimizer):
    """Wrapper class for generic optimizers/sparse regression methods

    Enables single target regressors (i.e. those whose predictions are
    1-dimensional) to perform multi target regression (i.e. predictions
    are 2-dimensional).  Also allows unbiasing & normalization for
    optimizers that would otherwise not include it.

    Args:
        optimizer: wrapped optimizer/sparse regression method

    Parameters
    ----------
    optimizer: estimator object
        The optimizer/sparse regressor to be wrapped, implementing ``fit`` and
        ``predict``. ``optimizer`` should also have the attribute ``coef_``.
        Any optimizer that supports a ``fit_intercept`` argument should
        be initialized to False.

    """

    def __init__(self, optimizer, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.optimizer = MultiOutputRegressor(optimizer)

    def _reduce(self, x, y):
        coef_shape = (y.shape[1], x.shape[1])
        self.coef_ = np.zeros(coef_shape)
        self.ind_ = np.ones(coef_shape)
        self.optimizer.fit(x, y)
        coef_list = [
            np.reshape(est.coef_, (-1, coef_shape[1]))
            for est in self.optimizer.estimators_
        ]
        self.coef_ = np.concatenate(coef_list, axis=0)
        self.ind_ = np.abs(self.coef_) > COEF_THRESHOLD
        self.intercept_ = 0.0
        return self

    def predict(self, x):
        return self.optimizer.predict(x)

    @property
    def complexity(self):
        return np.count_nonzero(self.coef_) + np.count_nonzero(self.intercept_)



"""
Base class for feature library classes.
"""
import abc
import warnings
from functools import wraps
from itertools import repeat
from typing import Optional
from typing import Sequence

import numpy as np
from scipy import sparse
from sklearn.base import BaseEstimator
from sklearn.base import TransformerMixin
from sklearn.utils.validation import check_is_fitted

from ..utils import AxesArray
from ..utils import comprehend_axes
from ..utils import validate_no_reshape
from ..utils import wrap_axes


class BaseFeatureLibrary(TransformerMixin, BaseEstimator):
    """
    Base class for feature libraries.

    Forces subclasses to implement ``fit``, ``transform``,
    and ``get_feature_names`` methods.
    """

    n_features_in_: int
    n_output_features_: int

    def validate_input(self, x, *args, **kwargs):
        return validate_no_reshape(x, *args, **kwargs)

    def reshape_samples_to_spatial_grid(self, x: np.ndarray) -> AxesArray:
        """Adapt predictions to fitted spatial grid."""
        spatial_grid = self.get_spatial_grid()
        if spatial_grid is None:
            return AxesArray(x, {"ax_sample": 0, "ax_coord": 1})
        # PDELibrary can have a zero spatial dimension for SINDyPI.
        shape = [dim for dim in self.get_spatial_grid().shape[:-1] if dim != 0]
        x = np.reshape(x, (*shape, -1, x.shape[-1]))
        return AxesArray(
            x,
            {
                "ax_spatial": list(range(len(shape))),
                "ax_sample": len(shape),
                "ax_coord": len(shape) + 1,
            },
        )

    def correct_shape(self, x: AxesArray):
        """Correct the shape of x, given what we know of the problem"""
        if len(x.shape) == 1:
            data = np.asarray(x).reshape((-1, 1))
            return AxesArray(data, {"ax_time": 0, "ax_coord": 1})
        elif len(x.shape) > 2 and type(self) is BaseFeatureLibrary:
            warnings.warn(
                "Data shapes with more than 2 axes are "
                "deprecated for the default problem.  We assume that time "
                "axis comes first, then coordinate axis, then all other "
                "axes continue the time axis.",
                DeprecationWarning,
            )
        return x

    def calc_trajectory(self, diff_method, x, t):
        x_dot = diff_method(x, t=t)
        x = AxesArray(diff_method.smoothed_x_, x.axes)
        return x, AxesArray(x_dot, x.axes)

    def get_spatial_grid(self):
        return None

    # Force subclasses to implement this
    @abc.abstractmethod
    def fit(self, x, y=None):
        """
        Compute number of output features.

        Parameters
        ----------
        x : array-like, shape (n_samples, n_features)
            The data.

        Returns
        -------
        self : instance
        """
        raise NotImplementedError

    # Force subclasses to implement this
    @abc.abstractmethod
    def transform(self, x):
        """
        Transform data.

        Parameters
        ----------
        x : array-like, shape [n_samples, n_features]
            The data to transform, row by row.

        Returns
        -------
        xp : np.ndarray, [n_samples, n_output_features]
            The matrix of features, where n_output_features is the number
            of features generated from the combination of inputs.
        """
        raise NotImplementedError

    # Force subclasses to implement this
    @abc.abstractmethod
    def get_feature_names(self, input_features=None) -> list[str]:
        """Return feature names for output features.

        Parameters
        ----------
        input_features : list of string, length n_features, optional
            String names for input features if available. By default,
            "x0", "x1", ... "xn_features" is used.

        Returns
        -------
        output_feature_names : list of string, length n_output_features
        """
        raise NotImplementedError

    def __add__(self, other):
        return ConcatLibrary([self, other])

    def __mul__(self, other):
        if isinstance(self, TensoredLibrary):
            return TensoredLibrary(self.libraries + [other])
        else:
            return TensoredLibrary([self, other])

    def __rmul__(self, other):
        return TensoredLibrary([self, other])

    @property
    def size(self):
        check_is_fitted(self)
        return self.n_output_features_


def x_sequence_or_item(wrapped_func):
    """Allow a feature library's method to handle list or item inputs."""

    @wraps(wrapped_func)
    def func(self, x, *args, **kwargs):
        if isinstance(x, Sequence):
            if isinstance(x[0], np.ndarray):
                xs = [AxesArray(xi, comprehend_axes(xi)) for xi in x]
            else:
                # e.g. jax array
                xs = x
            result = wrapped_func(self, xs, *args, **kwargs)
            # if transform() is a normal "return x"
            if isinstance(result, Sequence) and isinstance(result[0], np.ndarray):
                return [AxesArray(xp, comprehend_axes(xp)) for xp in result]
            return result  # e.g. fit() returns self
        else:
            if not sparse.issparse(x) and isinstance(x, np.ndarray):
                x = AxesArray(x, comprehend_axes(x))

                def reconstructor(x):
                    return x

            elif sparse.issparse(x):
                reconstructor = type(x)
                axes = comprehend_axes(x)
                wrap_axes(axes, x)
            else:  # e.g. jax array

                def reconstructor(x):
                    return x

            result = wrapped_func(self, [x], *args, **kwargs)
            if isinstance(result, Sequence):  # e.g. transform() returns x
                return reconstructor(result[0])
            return result  # e.g. fit() returns self

    return func


class ConcatLibrary(BaseFeatureLibrary):
    """Concatenate multiple libraries into one library. All settings
    provided to individual libraries will be applied.

    Parameters
    ----------
    libraries : list of libraries
        Library instances to be applied to the input matrix.

    Attributes
    ----------
    n_features_in_ : int
        The total number of input features.

    n_output_features_ : int
        The total number of output features. The number of output features
        is the sum of the numbers of output features for each of the
        concatenated libraries.

    Examples
    --------
    >>> import numpy as np
    >>> from pysindy.feature_library import FourierLibrary, CustomLibrary
    >>> from pysindy.feature_library import ConcatLibrary
    >>> x = np.array([[0.,-1],[1.,0.],[2.,-1.]])
    >>> functions = [lambda x : np.exp(x), lambda x,y : np.sin(x+y)]
    >>> lib_custom = CustomLibrary(library_functions=functions)
    >>> lib_fourier = FourierLibrary()
    >>> lib_concat = ConcatLibrary([lib_custom, lib_fourier])
    >>> lib_concat.fit()
    >>> lib.transform(x)
    """

    def __init__(
        self,
        libraries: list,
    ):
        self.libraries = libraries

    @x_sequence_or_item
    def fit(self, x_full, y=None):
        """
        Compute number of output features.

        Parameters
        ----------
        x : array-like, shape (n_samples, n_features)
            The data.

        Returns
        -------
        self : instance
        """
        n_features = x_full[0].shape[x_full[0].ax_coord]
        self.n_features_in_ = n_features

        # First fit all libs provided below
        fitted_libs = [lib.fit(x_full, y) for lib in self.libraries]

        # Calculate the sum of output features
        self.n_output_features_ = sum([lib.n_output_features_ for lib in fitted_libs])

        # Save fitted libs
        self.libraries = fitted_libs

        return self

    @x_sequence_or_item
    def transform(self, x_full):
        """Transform data with libs provided below.

        Parameters
        ----------
        x : array-like, shape [n_samples, n_features]
            The data to transform, row by row.

        Returns
        -------
        xp : np.ndarray, shape [n_samples, NP]
            The matrix of features, where NP is the number of features
            generated from applying the custom functions to the inputs.

        """
        for lib in self.libraries:
            check_is_fitted(lib)

        xp_full = []
        for x in x_full:
            feature_sets = [lib.transform([x])[0] for lib in self.libraries]
            xp = np.concatenate(feature_sets, axis=feature_sets[0].ax_coord)

            xp = AxesArray(xp, comprehend_axes(xp))
            xp_full.append(xp)
        return xp_full

    def get_feature_names(self, input_features=None):
        """Return feature names for output features.

        Parameters
        ----------
        input_features : list of string, length n_features, optional
            String names for input features if available. By default,
            "x0", "x1", ... "xn_features" is used.

        Returns
        -------
        output_feature_names : list of string, length n_output_features
        """
        feature_names = list()
        for lib in self.libraries:
            lib_feat_names = lib.get_feature_names(input_features)
            feature_names += lib_feat_names
        return feature_names

    def calc_trajectory(self, diff_method, x, t):
        return self.libraries[0].calc_trajectory(diff_method, x, t)


class TensoredLibrary(BaseFeatureLibrary):
    """Tensor multiple libraries together into one library. All settings
    provided to individual libraries will be applied.

    Parameters
    ----------
    libraries : list of libraries
        Library instances to be applied to the input matrix.

    inputs_per_library_ : Sequence of Sequences of ints (default None)
        list that specifies which input indexes should be passed as
        inputs for each of the individual feature libraries.
        length must equal the number of feature libraries.  Default is
        that all inputs are used for every library.

    Attributes
    ----------
    libraries_ : list of libraries
        Library instances to be applied to the input matrix.

    n_features_in_ : int
        The total number of input features.

    n_output_features_ : int
        The total number of output features. The number of output features
        is the product of the numbers of output features for each of the
        libraries that were tensored together.

    Examples
    --------
    >>> import numpy as np
    >>> from pysindy.feature_library import FourierLibrary, CustomLibrary
    >>> from pysindy.feature_library import TensoredLibrary
    >>> x = np.array([[0.,-1],[1.,0.],[2.,-1.]])
    >>> functions = [lambda x : np.exp(x), lambda x,y : np.sin(x+y)]
    >>> lib_custom = CustomLibrary(library_functions=functions)
    >>> lib_fourier = FourierLibrary()
    >>> lib_tensored = lib_custom * lib_fourier
    >>> lib_tensored.fit(x)
    >>> lib_tensored.transform(x)
    """

    def __init__(
        self,
        libraries: list,
        inputs_per_library: Optional[Sequence[Sequence[int]]] = None,
    ):
        self.libraries = libraries
        self.inputs_per_library = inputs_per_library

    def _combinations(self, lib_i: AxesArray, lib_j: AxesArray) -> AxesArray:
        """
        Compute combinations of the numerical libraries.

        Returns
        -------
        lib_full : All combinations of the numerical library terms.
        """
        # the shape here should be fixed with ax_coord....
        shape = np.array(lib_i.shape)
        shape[lib_i.ax_coord] = (
            lib_i.shape[lib_i.ax_coord] * lib_j.shape[lib_j.ax_coord]
        )
        lib_full = np.reshape(
            lib_i[..., :, "coord"] * lib_j[..., "coord", :],
            shape,
        )

        return lib_full

    def _name_combinations(self, lib_i, lib_j):
        """
        Compute combinations of the library feature names.

        Returns
        -------
        lib_full : All combinations of the library feature names.
        """
        lib_full = []
        for i in range(len(lib_i)):
            for j in range(len(lib_j)):
                lib_full.append(lib_i[i] + " " + lib_j[j])
        return lib_full

    def _set_inputs_per_library(self, inputs_per_library):
        """
        Extra function to make building a GeneralizedLibrary object easier
        """
        self.inputs_per_library = inputs_per_library

    @x_sequence_or_item
    def fit(self, x_full, y=None):
        """
        Compute number of output features.

        Parameters
        ----------
        x : array-like, shape (n_samples, n_features)
            The data.

        Returns
        -------
        self : instance
        """
        n_features = x_full[0].shape[x_full[0].ax_coord]

        self.n_features_in_ = n_features

        # If parameter is not set, use all the inputs
        if self.inputs_per_library is None:
            self.inputs_per_library = list(
                repeat(list(range(n_features)), len(self.libraries))
            )

        # First fit all libs provided below
        fitted_libs = [
            lib.fit([x[..., _unique(self.inputs_per_library[i])] for x in x_full], y)
            for i, lib in enumerate(self.libraries)
        ]

        # Calculate the sum of output features
        output_sizes = [lib.n_output_features_ for lib in fitted_libs]
        self.n_output_features_ = 1
        for osize in output_sizes:
            self.n_output_features_ *= osize

        # Save fitted libs
        self.libraries = fitted_libs

        return self

    @x_sequence_or_item
    def transform(self, x_full):
        """Transform data with libs provided below.

        Parameters
        ----------
        x : array-like, shape [n_samples, n_features]
            The data to transform, row by row.

        Returns
        -------
        xp : np.ndarray, shape [n_samples, NP]
            The matrix of features, where NP is the number of features
            generated from applying the custom functions to the inputs.

        """
        check_is_fitted(self)

        xp_full = []
        for x in x_full:
            xp = self.libraries[0].transform(
                [x[..., _unique(self.inputs_per_library[0])]]
            )[0]
            for i in range(1, len(self.libraries)):
                xp_i = self.libraries[i].transform(
                    [x[..., _unique(self.inputs_per_library[i])]]
                )[0]
                xp = self._combinations(xp, xp_i)

            xp_full.append(xp)
        return xp_full

    def get_feature_names(self, input_features: list[str] | None = None) -> list[str]:
        """Return feature names for output features.

        Parameters
        ----------
        input_features : list of string, length n_features, optional
            String names for input features if available. By default,
            "x0", "x1", ... "xn_features" is used.

        Returns
        -------
        output_feature_names : list of string, length n_output_features
        """
        check_is_fitted(self)

        if input_features is None:
            input_features = ["x%d" % i for i in range(self.n_features_in_)]

        feature_names = self.libraries[0].get_feature_names(
            np.asarray(input_features)[_unique(self.inputs_per_library[0])].tolist()
        )

        for i in range(1, len(self.libraries)):
            cur_input_features = np.asarray(input_features)[
                _unique(self.inputs_per_library[i])
            ].tolist()
            lib_i_feat_names = self.libraries[i].get_feature_names(cur_input_features)
            feature_names = self._name_combinations(feature_names, lib_i_feat_names)

        return feature_names

    def calc_trajectory(self, diff_method, x, t):
        return self.libraries[0].calc_trajectory(diff_method, x, t)


def _unique(s: Sequence) -> Sequence:
    """Remove duplicates, preserving order when python > 3.7"""
    return list(dict.fromkeys(s))



import numpy as np
import pytest
from numpy.testing import assert_
from numpy.testing import assert_array_equal
from numpy.testing import assert_equal
from numpy.testing import assert_raises

from pysindy import AxesArray
from pysindy.utils import _axes
from pysindy.utils._axes import _AxisMapping
from pysindy.utils._axes import AxesWarning


def test_axesarray_create():
    AxesArray(np.array(1), {})


def test_concat_out():
    arr = AxesArray(np.arange(3).reshape(1, 3), {"ax_a": 0, "ax_b": 1})
    arr_out = np.empty((2, 3)).view(AxesArray)
    result = np.concatenate((arr, arr), axis=0, out=arr_out)
    assert_equal(result, arr_out)


def test_bad_concat():
    arr = AxesArray(np.arange(3).reshape(1, 3), {"ax_a": 0, "ax_b": 1})
    arr2 = AxesArray(np.arange(3).reshape(1, 3), {"ax_b": 0, "ax_c": 1})
    with pytest.raises(ValueError):
        np.concatenate((arr, arr2), axis=0)


def test_reduce_mean_noinf_recursion():
    arr = AxesArray(np.array([[1]]), {"ax_a": [0, 1]})
    np.mean(arr, axis=0)


def test_repr():
    a = AxesArray(np.arange(5.0), {"ax_time": 0})
    result = a.__repr__()
    expected = "AxesArray([0., 1., 2., 3., 4.])"
    assert result == expected


def test_ufunc_override():
    # This is largely a clone of test_ufunc_override_with_super() from
    # numpy/core/tests/test_umath.py

    class B:
        def __array_ufunc__(self, ufunc, method, *inputs, **kwargs):
            if any(isinstance(input_, AxesArray) for input_ in inputs):
                return "A!"
            else:
                return NotImplemented

    d = np.arange(5.0)
    # 1 input, 1 output
    a = AxesArray(d, {"ax_time": 0})
    b = np.sin(a)
    check = np.sin(d)
    assert_(np.all(check == b))
    b = np.sin(d, out=(a,))
    assert_(np.all(check == b))
    assert_(b is a)
    a = AxesArray(np.arange(5.0), {"ax_time": 0})
    b = np.sin(a, out=a)
    assert_(np.all(check == b))

    # 1 input, 2 outputs
    a = AxesArray(np.arange(5.0), {"ax_time": 0})
    b1, b2 = np.modf(a)
    b1, b2 = np.modf(d, out=(None, a))
    assert_(b2 is a)
    a = AxesArray(np.arange(5.0), {"ax_time": 0})
    b = AxesArray(np.arange(5.0), {"ax_time": 0})
    c1, c2 = np.modf(a, out=(a, b))
    assert_(c1 is a)
    assert_(c2 is b)

    # 2 input, 1 output
    a = AxesArray(np.arange(5.0), {"ax_time": 0})
    b = AxesArray(np.arange(5.0), {"ax_time": 0})
    c = np.add(a, b, out=a)
    assert_(c is a)
    # some tests with a non-ndarray subclass
    a = np.arange(5.0)
    b = B()
    assert_(a.__array_ufunc__(np.add, "__call__", a, b) is NotImplemented)
    assert_(b.__array_ufunc__(np.add, "__call__", a, b) is NotImplemented)
    assert_raises(TypeError, np.add, a, b)
    a = AxesArray(a, {"ax_time": 0})
    assert_(a.__array_ufunc__(np.add, "__call__", a, b) is NotImplemented)
    assert_(b.__array_ufunc__(np.add, "__call__", a, b) == "A!")
    assert_(np.add(a, b) == "A!")
    # regression check for gh-9102 -- tests ufunc.reduce implicitly.
    d = np.array([[1, 2, 3], [1, 2, 3]])
    a = AxesArray(d, {"ax_time": [0, 1]})
    c = a.any()
    check = d.any()
    assert_equal(c, check)
    c = a.max()
    check = d.max()
    assert_equal(c, check)
    b = np.array(0).view(AxesArray)
    c = a.max(out=b)
    assert_equal(c, check)
    assert_(c is b)
    check = a.max(axis=0)
    b = np.zeros_like(check).view(AxesArray)
    c = a.max(axis=0, out=b)
    assert_equal(c, check)
    assert_(c is b)
    # simple explicit tests of reduce, accumulate, reduceat
    check = np.add.reduce(d, axis=1)
    c = np.add.reduce(a, axis=1)
    assert_equal(c, check)
    b = np.zeros_like(c)
    c = np.add.reduce(a, 1, None, b)
    assert_equal(c, check)
    assert_(c is b)
    check = np.add.accumulate(d, axis=0)
    c = np.add.accumulate(a, axis=0)
    assert_equal(c, check)
    b = np.zeros_like(c)
    c = np.add.accumulate(a, 0, None, b)
    assert_equal(c, check)
    assert_(c is b)
    indices = [0, 2, 1]
    check = np.add.reduceat(d, indices, axis=1)
    c = np.add.reduceat(a, indices, axis=1)
    assert_equal(c, check)
    b = np.zeros_like(c)
    c = np.add.reduceat(a, indices, 1, None, b)
    assert_equal(c, check)
    assert_(c is b)
    # and a few tests for at
    d = np.array([[1, 2, 3], [1, 2, 3]])
    check = d.copy()
    a = d.copy().view(AxesArray)
    np.add.at(check, ([0, 1], [0, 2]), 1.0)
    np.add.at(a, ([0, 1], [0, 2]), 1.0)
    assert_equal(np.asarray(a), np.asarray(check))  # modified
    b = np.array(1.0).view(AxesArray)
    a = d.copy().view(AxesArray)
    np.add.at(a, ([0, 1], [0, 2]), b)
    assert_equal(np.asarray(a), np.asarray(check))  # modified


def test_n_elements():
    arr = np.empty(np.arange(1, 5))
    arr = AxesArray(arr, {"ax_spatial": [0, 1], "ax_time": 2, "ax_coord": 3})
    assert arr.n_spatial == (1, 2)
    assert arr.n_time == 3
    assert arr.n_coord == 4

    arr2 = np.concatenate((arr, arr), axis=arr.ax_time)
    assert arr2.n_spatial == (1, 2)
    assert arr2.n_time == 6
    assert arr2.n_coord == 4


def test_reshape_outer_product():
    arr = AxesArray(np.arange(4).reshape((2, 2)), {"ax_a": [0, 1]})
    merge = np.reshape(arr, (4,))
    assert merge.axes == {"ax_a": 0}


def test_reshape_bad_divmod():
    arr = AxesArray(np.arange(12).reshape((2, 3, 2)), {"ax_a": [0, 1], "ax_b": 2})
    with pytest.raises(
        ValueError, match="Cannot reshape an AxesArray this way.  Array dimension"
    ):
        np.reshape(arr, (4, 3))


def test_reshape_fill_outer_product():
    arr = AxesArray(np.arange(4).reshape((2, 2)), {"ax_a": [0, 1]})
    merge = np.reshape(arr, (-1,))
    assert merge.axes == {"ax_a": 0}


def test_reshape_fill_regular():
    arr = AxesArray(np.arange(8).reshape((2, 2, 2)), {"ax_a": [0, 1], "ax_b": 2})
    merge = np.reshape(arr, (4, -1))
    assert merge.axes == {"ax_a": 0, "ax_b": 1}


def test_illegal_reshape():
    arr = AxesArray(np.arange(4).reshape((2, 2)), {"ax_a": [0, 1]})
    # melding across axes
    with pytest.raises(ValueError, match="Cannot reshape an AxesArray"):
        np.reshape(arr, (4, 1))

    # Add a hidden 1 in the middle!  maybe a matching 1

    # different name outer product
    arr = AxesArray(np.arange(4).reshape((2, 2)), {"ax_a": 0, "ax_b": 1})
    with pytest.raises(ValueError, match="Cannot reshape an AxesArray"):
        np.reshape(arr, (4,))
    # newaxes
    with pytest.raises(ValueError, match="Cannot reshape an AxesArray"):
        np.reshape(arr, (2, 1, 2))


def test_warn_toofew_axes():
    axes = {"ax_time": 0, "ax_coord": 1}
    with pytest.warns(AxesWarning):
        AxesArray(np.ones(8).reshape((2, 2, 2)), axes)


def test_toomany_axes():
    axes = {"ax_time": 0, "ax_coord": 2}
    with pytest.raises(ValueError):
        AxesArray(np.ones(4).reshape((2, 2)), axes)


def test_conflicting_axes_defn():
    axes = {"ax_time": 0, "ax_coord": 0}
    with pytest.raises(ValueError):
        AxesArray(np.ones(4), axes)


def test_missing_axis_errors():
    axes = {"ax_time": 0}
    arr = AxesArray(np.arange(3), axes)
    with pytest.raises(AttributeError):
        arr.ax_spatial
    with pytest.raises(AttributeError):
        arr.n_spatial


def test_simple_slice():
    arr = AxesArray(np.ones(2), {"ax_coord": 0})
    assert_array_equal(arr[:], arr)
    assert_array_equal(arr[slice(None)], arr)
    assert arr[0] == 1


# @pytest.mark.skip  # TODO: make this pass
def test_0d_indexer():
    arr = AxesArray(np.ones(2), {"ax_coord": 0})
    arr_out = arr[1, ...]
    assert arr_out.ndim == 0
    assert arr_out.axes == {}
    assert arr_out[()] == 1


def test_basic_indexing_modifies_axes():
    axes = {"ax_time": 0, "ax_coord": 1}
    arr = AxesArray(np.ones(4).reshape((2, 2)), axes)
    slim = arr[1, :, None]
    with pytest.raises(AttributeError):
        slim.ax_time
    assert slim.ax_unk == 1
    assert slim.ax_coord == 0
    reverse_slim = arr[None, :, 1]
    with pytest.raises(AttributeError):
        reverse_slim.ax_coord
    assert reverse_slim.ax_unk == 0
    assert reverse_slim.ax_time == 1
    almost_new = arr[None, None, 1, :, None, None]
    with pytest.raises(AttributeError):
        almost_new.ax_time
    assert almost_new.ax_coord == 2
    assert set(almost_new.ax_unk) == {0, 1, 3, 4}


def test_insert_named_axis():
    arr = AxesArray(np.ones(1), axes={"ax_time": 0})
    expanded = arr["time", :]
    result = expanded.axes
    expected = {"ax_time": [0, 1]}
    assert result == expected


def test_adv_indexing_modifies_axes():
    axes = {"ax_time": 0, "ax_coord": 1}
    arr = AxesArray(np.arange(4).reshape((2, 2)), axes)
    flat = arr[[0, 1], [0, 1]]
    same = arr[[[0], [1]], [0, 1]]
    tpose = arr[[0, 1], [[0], [1]]]
    assert flat.shape == (2,)
    np.testing.assert_array_equal(np.asarray(flat), np.array([0, 3]))

    assert flat.ax_time_coord == 0
    with pytest.raises(AttributeError):
        flat.ax_coord
    with pytest.raises(AttributeError):
        flat.ax_time

    assert same.shape == arr.shape
    np.testing.assert_equal(np.asarray(same), np.asarray(arr))
    assert same.ax_time_coord == [0, 1]
    with pytest.raises(AttributeError):
        same.ax_coord

    assert tpose.shape == arr.shape
    np.testing.assert_equal(np.asarray(tpose), np.asarray(arr.T))
    assert tpose.ax_time_coord == [0, 1]
    with pytest.raises(AttributeError):
        tpose.ax_coord


def test_adv_indexing_adds_axes():
    axes = {"ax_time": 0, "ax_coord": 1}
    arr = AxesArray(np.arange(4).reshape((2, 2)), axes)
    fat = arr[[[0, 1], [0, 1]]]
    assert fat.shape == (2, 2, 2)
    assert fat.ax_time == [0, 1]
    assert fat.ax_coord == 2


def test_standardize_basic_indexer():
    arr = np.arange(6).reshape(2, 3)
    result_indexer, result_fancy = _axes._standardize_indexer(arr, Ellipsis)
    assert result_indexer == [slice(None), slice(None)]
    assert result_fancy == ()

    result_indexer, result_fancy = _axes._standardize_indexer(
        arr, (np.newaxis, 1, 1, Ellipsis)
    )
    assert result_indexer == [None, 1, 1]
    assert result_fancy == ()


def test_standardize_advanced_indexer():
    arr = np.arange(6).reshape(2, 3)
    result_indexer, result_fancy = _axes._standardize_indexer(arr, [1])
    assert result_indexer == [np.ones(1), slice(None)]
    assert result_fancy == (0,)

    result_indexer, result_fancy = _axes._standardize_indexer(
        arr, (np.newaxis, [1], 1, Ellipsis)
    )
    assert result_indexer == [None, np.ones(1), 1]
    assert result_fancy == (1,)


def test_standardize_bool_indexer():
    arr = np.ones((1, 2))
    result, result_adv = _axes._standardize_indexer(arr, [[True, True]])
    assert_equal(result, [[0, 0], [0, 1]])
    assert result_adv == (0, 1)


def test_reduce_AxisMapping():
    ax_map = _AxisMapping(
        {"ax_a": [0, 1], "ax_b": 2, "ax_c": 3, "ax_d": 4, "ax_e": [5, 6]},
        7,
    )
    result = ax_map.remove_axis(3)
    expected = {"ax_a": [0, 1], "ax_b": 2, "ax_d": 3, "ax_e": [4, 5]}
    assert result == expected
    result = ax_map.remove_axis(-4)
    assert result == expected


def test_reduce_all_AxisMapping():
    ax_map = _AxisMapping({"ax_a": [0, 1], "ax_b": 2}, 3)
    result = ax_map.remove_axis()
    expected = {}
    assert result == expected


def test_reduce_multiple_AxisMapping():
    ax_map = _AxisMapping(
        {
            "ax_a": [0, 1],
            "ax_b": 2,
            "ax_c": 3,
            "ax_d": 4,
            "ax_e": [5, 6],
        },
        7,
    )
    result = ax_map.remove_axis([3, 4])
    expected = {
        "ax_a": [0, 1],
        "ax_b": 2,
        "ax_e": [3, 4],
    }
    assert result == expected


def test_reduce_twisted_AxisMapping():
    ax_map = _AxisMapping(
        {
            "ax_a": [0, 6],
            "ax_b": 2,
            "ax_c": 3,
            "ax_d": 4,
            "ax_e": [1, 5],
        },
        7,
    )
    result = ax_map.remove_axis([3, 4])
    expected = {
        "ax_a": [0, 4],
        "ax_b": 2,
        "ax_e": [1, 3],
    }
    assert result == expected


def test_reduce_misordered_AxisMapping():
    ax_map = _AxisMapping({"ax_a": [0, 1], "ax_b": 2, "ax_c": 3}, 4)
    result = ax_map.remove_axis([2, 1])
    expected = {"ax_a": 0, "ax_c": 1}
    assert result == expected


def test_insert_AxisMapping():
    ax_map = _AxisMapping(
        {
            "ax_a": [0, 1],
            "ax_b": 2,
            "ax_c": 3,
            "ax_d": [4, 5],
        },
        6,
    )
    result = ax_map.insert_axis(3, "ax_unk")
    expected = {
        "ax_a": [0, 1],
        "ax_b": 2,
        "ax_unk": 3,
        "ax_c": 4,
        "ax_d": [5, 6],
    }
    assert result == expected


def test_insert_existing_AxisMapping():
    ax_map = _AxisMapping(
        {
            "ax_a": [0, 1],
            "ax_b": 2,
            "ax_c": 3,
            "ax_d": [4, 5],
        },
        6,
    )
    result = ax_map.insert_axis(3, "ax_b")
    expected = {
        "ax_a": [0, 1],
        "ax_b": [2, 3],
        "ax_c": 4,
        "ax_d": [5, 6],
    }
    assert result == expected


def test_insert_multiple_AxisMapping():
    ax_map = _AxisMapping(
        {
            "ax_a": [0, 1],
            "ax_b": 2,
            "ax_c": 3,
            "ax_d": [4, 5],
        },
        6,
    )
    result = ax_map.insert_axis([1, 4], new_name="ax_unk")
    expected = {
        "ax_a": [0, 2],
        "ax_unk": [1, 4],
        "ax_b": 3,
        "ax_c": 5,
        "ax_d": [6, 7],
    }
    assert result == expected


def test_insert_misordered_AxisMapping():
    ax_map = _AxisMapping(
        {
            "ax_a": [0, 1],
            "ax_b": 2,
            "ax_c": 3,
            "ax_d": [4, 5],
        },
        6,
    )
    result = ax_map.insert_axis([4, 1], new_name="ax_unk")
    expected = {
        "ax_a": [0, 2],
        "ax_unk": [1, 4],
        "ax_b": 3,
        "ax_c": 5,
        "ax_d": [6, 7],
    }
    assert result == expected


def test_determine_adv_broadcasting():
    indexers = (1, np.ones(1), np.ones((4, 1)), np.ones(3))
    res_nd, res_start = _axes._determine_adv_broadcasting(indexers, [1, 2, 3])
    assert res_nd == 2
    assert res_start == 1

    indexers = (None, np.ones(1), 2, np.ones(3))
    res_nd, res_start = _axes._determine_adv_broadcasting(indexers, [1, 3])
    assert res_nd == 1
    assert res_start == 0

    res_nd, res_start = _axes._determine_adv_broadcasting(indexers, [])
    assert res_nd == 0
    assert res_start is None


def test_replace_ellipsis():
    key = [..., 0]
    result = _axes._expand_indexer_ellipsis(key, 2)
    expected = [slice(None), 0]
    assert result == expected


def test_strip_ellipsis():
    key = [1, ...]
    result = _axes._expand_indexer_ellipsis(key, 1)
    expected = [1]
    assert result == expected

    key = [..., 1]
    result = _axes._expand_indexer_ellipsis(key, 1)
    expected = [1]
    assert result == expected


def test_transpose():
    axes = {"ax_a": 0, "ax_b": [1, 2]}
    arr = AxesArray(np.arange(8).reshape(2, 2, 2), axes)
    tp = np.transpose(arr, [2, 0, 1])
    result = tp.axes
    expected = {"ax_a": 1, "ax_b": [0, 2]}
    assert result == expected
    assert_array_equal(tp, np.transpose(np.asarray(arr), [2, 0, 1]))
    arr = arr[..., 0]
    tp = arr.T
    expected = {"ax_a": 1, "ax_b": 0}
    assert_array_equal(tp, np.asarray(arr).T)


def test_linalg_solve_align_left():
    axesA = {"ax_prob": 0, "ax_sample": 1, "ax_coord": 2}
    arrA = AxesArray(np.arange(8).reshape(2, 2, 2), axesA)
    axesb = {"ax_sample": 0, "ax_feat": 1}
    arrb = AxesArray(np.arange(4).reshape(2, 2), axesb)
    result = np.linalg.solve(arrA, arrb)
    expected_axes = {"ax_prob": 0, "ax_coord": 1, "ax_feat": 2}
    assert result.axes == expected_axes
    super_result = np.linalg.solve(np.asarray(arrA), np.asarray(arrb))
    assert_array_equal(result, super_result)


def test_linalg_solve_align_right():
    axesA = {"ax_sample": 0, "ax_feature": 1}
    arrA = AxesArray(np.arange(4).reshape(2, 2), axesA)
    axesb = {"ax_sample": 0, "ax_target": 1}
    arrb = AxesArray(np.arange(4).reshape(2, 2), axesb)
    result = np.linalg.solve(arrA, arrb)
    expected_axes = {"ax_feature": 0, "ax_target": 1}
    assert result.axes == expected_axes
    super_result = np.linalg.solve(np.asarray(arrA), np.asarray(arrb))
    assert_array_equal(result, super_result)


def test_linalg_solve_align_right_xl():
    axesA = {"ax_sample": 0, "ax_feature": 1}
    arrA = AxesArray(np.arange(4).reshape(2, 2), axesA)
    axesb = {"ax_prob": 0, "ax_sample": 1, "ax_target": 2}
    arrb = AxesArray(np.arange(8).reshape(2, 2, 2), axesb)
    result = np.linalg.solve(arrA, arrb)
    expected_axes = {"ax_prob": 0, "ax_feature": 1, "ax_target": 2}
    assert result.axes == expected_axes
    super_result = np.linalg.solve(np.asarray(arrA), np.asarray(arrb))
    assert_array_equal(result, super_result)


def test_linalg_solve_incompatible_left():
    axesA = {"ax_prob": 0, "ax_sample": 1, "ax_coord": 2}
    arrA = AxesArray(np.arange(8).reshape(2, 2, 2), axesA)
    axesb = {"ax_foo": 0, "ax_sample": 1}
    arrb = AxesArray(np.arange(4).reshape(2, 2), axesb)
    with pytest.raises(ValueError, match="Mismatch in operand axis names"):
        np.linalg.solve(arrA, arrb)


def test_ts_to_einsum_int_axes():
    a_str, b_str = _axes._tensordot_to_einsum(3, 3, 2).split(",")
    # expecting 'abc,bcf
    assert a_str[0] not in b_str
    assert b_str[-1] not in a_str
    assert a_str[1:] == b_str[:-1]


def test_ts_to_einsum_list_axes():
    a_str, b_str = _axes._tensordot_to_einsum(3, 3, [[1], [2]]).split(",")
    # expecting 'abcd,efbh
    assert a_str[1] == b_str[2]
    assert a_str[0] not in b_str
    assert a_str[2] not in b_str
    assert b_str[0] not in a_str
    assert b_str[1] not in a_str


def test_tensordot_int_axes():
    axes_a = {"ax_a": 0, "ax_b": [1, 2]}
    axes_b = {"ax_b": [0, 1], "ax_c": 2}
    arr = np.arange(8).reshape((2, 2, 2))
    arr_a = AxesArray(arr, axes_a)
    arr_b = AxesArray(arr, axes_b)
    super_result = np.tensordot(arr, arr, 2)
    result = np.tensordot(arr_a, arr_b, 2)
    expected_axes = {"ax_a": 0, "ax_c": 1}
    assert result.axes == expected_axes
    assert_array_equal(result, super_result)


def test_tensordot_list_axes():
    axes_a = {"ax_a": 0, "ax_b": [1, 2]}
    axes_b = {"ax_c": [0, 1], "ax_b": 2}
    arr = np.arange(8).reshape((2, 2, 2))
    arr_a = AxesArray(arr, axes_a)
    arr_b = AxesArray(arr, axes_b)
    super_result = np.tensordot(arr, arr, [[1], [2]])
    result = np.tensordot(arr_a, arr_b, [[1], [2]])
    expected_axes = {"ax_a": 0, "ax_b": 1, "ax_c": [2, 3]}
    assert result.axes == expected_axes
    assert_array_equal(result, super_result)


def test_ravel_1d():
    arr = AxesArray(np.array([1, 2]), axes={"ax_a": 0})
    result = np.ravel(arr)
    assert_array_equal(result, arr)
    assert result.axes == arr.axes


def test_ravel_nd():
    arr = AxesArray(np.array([[1, 2], [3, 4]]), axes={"ax_a": 0, "ax_b": 1})
    result = np.ravel(arr)
    expected = np.ravel(np.asarray(arr))
    assert_array_equal(result, expected)
    assert result.axes == {"ax_unk": 0}


def test_ma_ravel():
    arr = AxesArray(np.array([1, 2]), axes={"ax_a": 0})
    marr = np.ma.MaskedArray(arr)
    np.ma.ravel(marr)


@pytest.mark.skip
def test_einsum_implicit():
    ...


@pytest.mark.skip
def test_einsum_trace():
    ...


@pytest.mark.skip
def test_einsum_diag():
    ...


@pytest.mark.skip
def test_einsum_1dsum():
    ...


@pytest.mark.skip
def test_einsum_alldsum():
    ...


@pytest.mark.skip
def test_einsum_contraction():
    ...


@pytest.mark.skip
def test_einsum_explicit_ellipsis():
    ...


def test_einsum_scalar():
    arr = AxesArray(np.ones(1), {"ax_a": 0})
    expected = 1
    result = np.einsum("i,i", arr, arr)
    assert result == expected


@pytest.mark.skip
def test_einsum_mixed():
    ...



import pytest
from hypothesis import assume
from hypothesis import given
from hypothesis import settings
from hypothesis.strategies import integers
from sklearn.datasets import make_regression
from sklearn.linear_model import Lasso
from sklearn.linear_model import Ridge

from pysindy.optimizers import SR3
from pysindy.optimizers import STLSQ
from pysindy.optimizers import WrappedOptimizer


@pytest.mark.skip
@pytest.mark.parametrize(
    "opt_cls, reg_name",
    [[Lasso, "alpha"], [Ridge, "alpha"], [STLSQ, "threshold"], [SR3, "threshold"]],
)
@given(
    n_samples=integers(min_value=30, max_value=50),
    n_features=integers(min_value=6, max_value=10),
    n_informative=integers(min_value=2, max_value=5),
    random_state=integers(min_value=0, max_value=2**32 - 1),
)
@settings(max_examples=20, deadline=None)
def test_complexity_parameter(
    opt_cls, reg_name, n_samples, n_features, n_informative, random_state
):
    """Behaviour test for complexity 2.

    We assume that a model with a bigger regularization parameter is less complex.
    """
    assume(n_informative <= n_features)

    x, y = make_regression(
        n_samples=n_samples,
        n_features=n_features,
        n_informative=n_informative,
        n_targets=1,
        bias=0,
        noise=0.1,
        random_state=random_state,
    )
    y = y.reshape(-1, 1)

    optimizers = [
        WrappedOptimizer(opt_cls(**{reg_name: reg_value}), normalize_columns=True)
        for reg_value in [10, 1, 0.1, 0.001]
    ]

    for opt in optimizers:
        opt.fit(x, y)

    for less_complex, more_complex in zip(optimizers, optimizers[1:]):
        assert less_complex.complexity <= more_complex.complexity



import numpy as np
import pytest
import scipy

from pysindy import PolynomialLibrary
from pysindy import TrappingSR3
from pysindy.optimizers.trapping_sr3 import _antisymm_double_constraint
from pysindy.optimizers.trapping_sr3 import _antisymm_triple_constraints
from pysindy.optimizers.trapping_sr3 import _make_constraints


@pytest.fixture
def poly_lib_terms_coef_bias(scope="session"):
    lib = PolynomialLibrary(2, include_bias=True).fit(np.zeros((1, 2)))
    # terms are [1, x, y, x^2 , xy, y^2]
    polyterms = [(t_ind, exps) for t_ind, exps in enumerate(lib.powers_)]
    coeffs = np.array([[1, 2, 3, 4, 5, 6], [7, 8, 9, 10, 11, 12]])
    return lib, polyterms, coeffs


@pytest.fixture
def poly_lib_terms_coef_nobias(scope="session"):
    lib = PolynomialLibrary(2, include_bias=False).fit(np.zeros((1, 2)))
    # terms are [x, y, x^2 , xy, y^2]
    polyterms = [(t_ind, exps) for t_ind, exps in enumerate(lib.powers_)]
    coeffs = np.array([[2, 3, 4, 5, 6], [8, 9, 10, 11, 12]])
    return lib, polyterms, coeffs


@pytest.mark.parametrize(
    "lib_terms_coeffs",
    (
        pytest.lazy_fixture("poly_lib_terms_coef_bias"),
        pytest.lazy_fixture("poly_lib_terms_coef_nobias"),
    ),
)
def test_PL(lib_terms_coeffs):
    _, terms, coeffs = lib_terms_coeffs
    PL_symm, PL_unsymm = TrappingSR3._build_PL(terms)

    expected_symm = np.array([[2.0, 5.5], [5.5, 9.0]])
    expected_unsymm = np.array([[2.0, 3.0], [8.0, 9.0]])
    result = np.einsum("ijkl,kl", PL_symm, coeffs)
    np.testing.assert_array_equal(result, expected_symm)
    result = np.einsum("ijkl,kl", PL_unsymm, coeffs)
    np.testing.assert_array_equal(result, expected_unsymm)


@pytest.mark.parametrize(
    "lib_terms_coeffs",
    (
        pytest.lazy_fixture("poly_lib_terms_coef_bias"),
        pytest.lazy_fixture("poly_lib_terms_coef_nobias"),
    ),
)
def test_PQ(lib_terms_coeffs):
    _, terms, coeffs = lib_terms_coeffs
    PQ = TrappingSR3._build_PQ(terms)
    expected = np.array([[[4.0, 2.5], [2.5, 6]], [[10.0, 5.5], [5.5, 12.0]]])
    result = np.einsum("ijklm,lm", PQ, coeffs)
    np.testing.assert_array_equal(result, expected)


def test_enstrophy_constraints_imply_enstrophy_symmetry():
    n_tgts = 4
    root = np.random.normal(size=(n_tgts, n_tgts))
    mod_matrix = root @ root.T
    bias = False
    lib = PolynomialLibrary(2, include_bias=bias).fit(np.ones((1, n_tgts)))
    terms = [(t_ind, exps) for t_ind, exps in enumerate(lib.powers_)]
    PQ = TrappingSR3._build_PQ(terms)

    _, constraint_lhs = _make_constraints(n_tgts, include_bias=bias)
    constraint_lhs = np.tensordot(constraint_lhs, mod_matrix, axes=1)
    n_constraint, n_features, _ = constraint_lhs.shape
    constraint_mat = constraint_lhs.reshape((n_constraint, -1))
    coeff_basis = scipy.linalg.null_space(constraint_mat)
    _, constraint_nullity = coeff_basis.shape
    coeffs = coeff_basis @ np.random.normal(size=(constraint_nullity, 1))
    coeffs = coeffs.reshape((n_features, n_tgts))

    Q = np.tensordot(PQ, coeffs, axes=([4, 3], [0, 1]))
    Q_tilde = np.tensordot(mod_matrix, Q, axes=1)
    Q_permsum = (
        Q_tilde + np.transpose(Q_tilde, [1, 2, 0]) + np.transpose(Q_tilde, [2, 0, 1])
    )
    np.testing.assert_allclose(np.zeros_like(Q_permsum), Q_permsum, atol=1e-14)


def test_enstrophy_symmetry_implies_enstrophy_constraints():
    n_tgts = 4
    root = np.random.normal(size=(n_tgts, n_tgts))
    mod_matrix = root @ root.T
    u, _, vt = np.linalg.svd(mod_matrix)
    mod_matrix = u @ vt
    mod_inv = np.linalg.inv(mod_matrix)
    bias = False
    lib = PolynomialLibrary(2, include_bias=bias).fit(np.ones((1, n_tgts)))
    terms = [(t_ind, exps) for t_ind, exps in enumerate(lib.powers_)]
    PQ = TrappingSR3._build_PQ(terms)
    PQinv = np.zeros_like(PQ)
    PQinv[np.where(PQ != 0)] = 1

    Q_tilde = np.random.normal(size=(n_tgts, n_tgts, n_tgts))
    Q_tilde[(range(n_tgts),) * 3] = 0
    Q_tilde = (Q_tilde + np.transpose(Q_tilde, [0, 2, 1])) / 2
    Q_tilde -= (
        Q_tilde + np.transpose(Q_tilde, [1, 2, 0]) + np.transpose(Q_tilde, [2, 0, 1])
    ) / 3
    # Assert symmetry
    Qperm = (
        Q_tilde + np.transpose(Q_tilde, [1, 2, 0]) + np.transpose(Q_tilde, [2, 0, 1])
    )
    np.testing.assert_allclose(Qperm, np.zeros_like(Qperm), atol=1e-15)
    Q = np.tensordot(mod_inv, Q_tilde, axes=1)

    # transpose:  make_constraints is (features, targets), but PQ is (targets, features)
    coeffs = np.tensordot(PQinv, Q, axes=([0, 1, 2], [0, 1, 2])).T
    expected, constraint_lhs = _make_constraints(n_tgts, include_bias=bias)
    constraint_lhs = np.tensordot(constraint_lhs, mod_matrix, axes=1)
    n_constraints, _, _ = constraint_lhs.shape
    result = constraint_lhs.reshape((n_constraints, -1)) @ coeffs.flatten()
    np.testing.assert_allclose(result, expected, atol=1e-15)


def test_constraints_imply_symmetry():
    n_tgts = 4
    bias = False
    lib = PolynomialLibrary(2, include_bias=bias).fit(np.ones((1, n_tgts)))
    terms = [(t_ind, exps) for t_ind, exps in enumerate(lib.powers_)]
    PQ = TrappingSR3._build_PQ(terms)

    _, constraint_lhs = _make_constraints(n_tgts, include_bias=bias)
    n_constraint, n_features, _ = constraint_lhs.shape
    constraint_mat = constraint_lhs.reshape((n_constraint, -1))
    coeff_basis = scipy.linalg.null_space(constraint_mat)
    _, constraint_nullity = coeff_basis.shape
    coeffs = coeff_basis @ np.random.normal(size=(constraint_nullity, 1))
    coeffs = coeffs.reshape((n_features, n_tgts))

    Q = np.tensordot(PQ, coeffs, axes=([4, 3], [0, 1]))
    Q_permsum = Q + np.transpose(Q, [1, 2, 0]) + np.transpose(Q, [2, 0, 1])
    np.testing.assert_allclose(np.zeros_like(Q_permsum), Q_permsum, atol=1e-15)


def test_symmetry_implies_constraints():
    n_tgts = 4
    bias = False
    lib = PolynomialLibrary(2, include_bias=bias).fit(np.ones((1, n_tgts)))
    terms = [(t_ind, exps) for t_ind, exps in enumerate(lib.powers_)]
    PQ = TrappingSR3._build_PQ(terms)
    PQinv = np.zeros_like(PQ)
    PQinv[np.where(PQ != 0)] = 1

    Q = np.random.normal(size=(n_tgts, n_tgts, n_tgts))
    Q[(range(n_tgts),) * 3] = 0
    Q = (Q + np.transpose(Q, [0, 2, 1])) / 2
    Q -= (Q + np.transpose(Q, [1, 2, 0]) + np.transpose(Q, [2, 0, 1])) / 3
    # Assert symmetry
    Qperm = Q + np.transpose(Q, [1, 2, 0]) + np.transpose(Q, [2, 0, 1])
    np.testing.assert_allclose(Qperm, np.zeros_like(Qperm), atol=1e-15)

    # transpose:  make_constraints is (features, targets), but PQ is (targets, features)
    coeffs = np.tensordot(PQinv, Q, axes=([0, 1, 2], [0, 1, 2])).T
    expected, constraint_lhs = _make_constraints(n_tgts, include_bias=bias)
    n_constraints, _, _ = constraint_lhs.shape
    result = constraint_lhs.reshape((n_constraints, -1)) @ coeffs.flatten()
    np.testing.assert_allclose(result, expected, atol=1e-15)


@pytest.mark.parametrize("include_bias", (True, False))
def test_trapping_constraints(include_bias):
    # x, y, x^2, xy, y^2
    constraint_rhs, constraint_lhs = _make_constraints(2, include_bias=include_bias)
    stable_coefs = np.array([[0, 0, 0, 1, -1], [0, 0, -1, 1, 0]])
    if include_bias:
        stable_coefs = np.concatenate(([[0], [0]], stable_coefs), axis=1)
    result = np.tensordot(constraint_lhs, stable_coefs, ((1, 2), (1, 0)))
    np.testing.assert_array_equal(constraint_rhs, result)
    _, lg_constraint = _make_constraints(4, include_bias=include_bias)
    # constraint should be full-rank
    expected = len(lg_constraint)
    result = np.linalg.matrix_rank(lg_constraint.reshape((expected, -1)))
    assert result == expected


def test_trapping_triple_mixed_constraint():
    # xy, xz, yz
    stable_coefs = np.array([[0, 0, -1], [0, 0.5, 0], [0.5, 0, 0]])
    mixed_terms = {frozenset((0, 1)): 0, frozenset((0, 2)): 1, frozenset((1, 2)): 2}
    constraint_lhs = _antisymm_triple_constraints(3, 3, mixed_terms)
    result = np.tensordot(constraint_lhs, stable_coefs, ((1, 2), (1, 0)))
    np.testing.assert_array_equal(result, np.zeros_like(result))


def test_trapping_double_constraint():
    stable_coefs = np.array(
        [
            # w^2, wx, wy, wz, x^2, xy, xz, y^2, yz, z^2
            [0, 1, 2, 3, -4, 0, 0, -8, 0, -9],  # w
            [-1, 4, 0, 0, 0, 5, 6, -10, 0, -11],  # x
            [-2, 0, 8, 0, -5, 10, 0, 0, 7, -12],  # y
            [-3, 0, 0, 9, -6, 0, 11, -7, 12, 0],  # z
        ]
    )
    pure_terms = {0: 0, 1: 4, 2: 7, 3: 9}
    mixed_terms = {
        frozenset((0, 1)): 1,
        frozenset((0, 2)): 2,
        frozenset((0, 3)): 3,
        frozenset((1, 2)): 5,
        frozenset((1, 3)): 6,
        frozenset((2, 3)): 8,
    }
    constraint_lhs = _antisymm_double_constraint(4, 10, pure_terms, mixed_terms)
    result = np.tensordot(constraint_lhs, stable_coefs, ((1, 2), (1, 0)))
    np.testing.assert_array_equal(result, np.zeros_like(result))
    # constraint should be full-rank
    expected = len(constraint_lhs)
    result = np.linalg.matrix_rank(constraint_lhs.reshape((expected, -1)))
    assert result == expected
